# -*- coding: utf-8 -*-
# BFit is a Python library for fitting a convex sum of Gaussian
# functions to any probability distribution
#
# Copyright (C) 2020- The QC-Devs Community
#
# This file is part of BFit.
#
# BFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# BFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# ---
r"""Parsing Slater files."""
import os
import re

import numpy as np


__all__ = ["load_slater_wfn"]


def load_slater_wfn(element, anion=False, cation=False):
    """
    Return the data inside the atomic Slater '.slater' files wave-function file as a dictionary.

    Parameters
    ----------
    element : str
        The atom/element.
    anion : bool
        If true, then the anion of element is used.
    cation : bool
        If true, then the cation of element is used.
    """
    # Heavy atoms from atom cs to lr.
    heavy_atoms = ["cs", "ba", "la", "ce", "pr", "nd", "pm", "sm", "eu", "gd", "tb", "dy", "ho",
                   "er", "tm", "yb", "lu", "hf", "ta", "w", "re", "os", "ir", "pt", "au", "hg",
                   "tl", "pb", "bi", "po", "at", "rn", "fr", "ra", "ac", "th", "pa", "u", "np",
                   "pu", "am", "cm", "bk", "cf", "es", "fm", "md", "no", "lr"]

    anion_atoms = ["ag", "al", "as", "b", "br", "c", "cl", "co", "cr", "cu", "f", "fe", "ga",
                   "ge", "h", "i", "in", "k", "li", "mn", "mo", "n", "na", "nb", "ni", "o",
                   "p", "pd", "rb", "rh", "ru", "s", "sb", "sc", "se", "si", "sn", "tc", "te",
                   "ti", "v", "y", "zr"]

    cation_atoms = ["ag", "al", "ar", "as", "b", "be", "br", "c", "ca", "cd", "cl", "co",
                    "cr", "cs", "cu", "f", "fe", "ga", "ge", "i", "in", "k", "kr", "li",
                    "mg", "mn", "mo", "n", "na", "nb", "ne", "ni", "o", "p", "pd", "rb",
                    "rh", "ru", "s", "sb", "sc", "se", "si", "sn", "sr", "tc", "te", "ti",
                    "v", "xe", "y", "zn", "zr"]

    is_heavy_element = element.lower() in heavy_atoms
    if anion:
        if element.lower() in anion_atoms:
            file_path = f"/data/anion/{element.lower()}.an"
        else:
            raise ValueError(
                f"Anion Slater File for element {element} does not exist."
            )
    elif cation:
        if element.lower() in cation_atoms:
            file_path = f"/data/cation/{element.lower()}.cat"
        else:
            raise ValueError(
                f"Cation Slater File for element {element} does not exist."
            )
    else:
        file_path = f"/data/neutral/{element.lower()}.slater"

    file_name = os.path.join(os.path.dirname(__file__) + file_path)

    def _get_number_of_electrons_per_orbital(configuration):
        """
        Get the Occupation Number for all orbitals of an _element returing an dictionary.

        Parameters
        ----------
        configuration : str
            The electron configuration.

        Returns
        --------
        dict
            a dict containing the number and orbital.

        """
        electron_config_list = configuration

        shells = ["K", "L", "M", "N"]

        out = {}
        orbitals = [str(x) + "S" for x in range(1, 8)] + [str(x) + "P" for x in range(2, 8)] + \
                   [str(x) + "D" for x in range(3, 8)] + [str(x) + "F" for x in range(4, 8)]
        for orb in orbitals:
            # Initialize all atomic orbitals to zero electrons
            out[orb] = 0

        # Sometimes electron configuration includes K L M N for simplicity
        for x in shells:
            if x in electron_config_list:
                if x == "K":
                    out["1S"] = 2
                elif x == "L":
                    out["2S"] = 2
                    out["2P"] = 6
                elif x == "M":
                    out["3S"] = 2
                    out["3P"] = 6
                    out["3D"] = 10
                elif x == "N":
                    out["4S"] = 2
                    out["4P"] = 6
                    out["4D"] = 10
                    out["4F"] = 14

        for x in orbitals:
            if x in electron_config_list:
                index = electron_config_list.index(x)
                orbital = (electron_config_list[index: index + 2])

                if orbital[1] == "D" or orbital[1] == "F":
                    num_electrons = re.search(orbital + r"\((.*?)\)", electron_config_list).group(1)
                    out[orbital] = int(num_electrons)
                else:
                    out[orbital] = int(electron_config_list[index + 3: index + 4])

        return {key: value for key, value in out.items() if value != 0}

    def _get_column(t_orbital):
        """
        Correct the error in order to retrieve the correct column.

        The Columns are harder to parse as the orbitals start with one while p orbitals start at
            energy.

        Parameters
        ----------
        t_orbital : str
            orbital i.e. "1S" or "2P" or "3D"

        Returns
        -------
        int :
            Retrieve the right column index depending on whether it is "S", "P" or "D" orbital.

        """
        if t_orbital[1] == "S":
            return int(t_orbital[0]) + 1
        elif t_orbital[1] == "P":
            return int(t_orbital[0])
        elif t_orbital[1] == "D":
            return int(t_orbital[0]) - 1
        elif t_orbital[1] == "F":
            return int(t_orbital[0]) - 2
        else:
            raise ValueError(f"Did not recognize orbital {t_orbital}.")

    def _configuration_exact_for_heavy_elements(configuration):
        r"""Later file for heavy elements does not contain the configuration in right format."""
        true_configuration = ""
        if "[XE]" in configuration:
            true_configuration += "K(2)L(8)M(18)4S(2)4P(6)5S(2)4D(10)5P(6)"
            true_configuration += configuration.split("[XE]")[1]
        elif "[RN]" in configuration:
            # Add Xenon
            true_configuration += "K(2)L(8)M(18)4S(2)4P(6)5S(2)4D(10)5P(6)"
            # Add Rn
            true_configuration += "4F(14)6S(2)5D(10)6P(6)"
            # Add rest
            true_configuration += configuration.split("[RN]")[1]
        else:
            raise ValueError("Heavy element is not the right format for parsing.")
        return true_configuration

    with open(file_name, "r", encoding="utf8") as f:
        line = f.readline()
        configuration = line.split()[1].replace(",", "")
        if is_heavy_element:
            configuration = _configuration_exact_for_heavy_elements(configuration)

        next_line = f.readline()
        # Sometimes there are blank lines.
        while len(next_line.strip()) == 0:
            next_line = f.readline()
        if is_heavy_element:
            # Heavy element slater files has extra redundant information of 5 lines.
            for _ in range(0, 6):
                next_line = f.readline()

        # Get energy from "E=..." line
        split_energy_line = next_line.split("=")
        if not split_energy_line[0].strip() == "E":
            raise RuntimeError("Parsing error of energy term 'E='.")
        energy = float(split_energy_line[1])

        # Split the kinetic, potential energy.
        split_energy_line = re.findall(r"[= -]\d+.\d+", f.readline())
        assert len(split_energy_line) == 3
        kinetic_energy = float(split_energy_line[0])
        potential_energy = float(split_energy_line[1])

        orbitals = []
        orbitals_basis = {'S': [], 'P': [], 'D': [], "F": []}
        orbitals_cusp = []
        orbitals_energy = []
        orbitals_exp = {'S': [], 'P': [], 'D': [], "F": []}
        orbitals_coeff = {}

        line = f.readline()
        while line.strip() == "":
            line = f.readline()
        while line.strip() != "":
            # If line has ___S___ or P or D where _ = " ".
            if re.search(r'  [S|P|D|F]  ', line):
                # Get All The Orbitals
                subshell = line.split()[0]
                list_of_orbitals = line.split()[1:]
                orbitals += list_of_orbitals
                for x in list_of_orbitals:
                    orbitals_coeff[x] = []   # Initilize orbitals inside coefficient dictionary

                # Get Energy, Cusp Levels
                line = f.readline()
                orbitals_energy.extend([float(x) for x in line.split()[1:]])
                if not is_heavy_element:
                    # Heavy atoms slater files, doesn't have cusp values,.
                    line = f.readline()
                    orbitals_cusp.extend([float(x) for x in line.split()[1:]])
                line = f.readline()

                # Get Exponents, Coefficients, Orbital Basis
                while re.match(r'\A^\d' + subshell, line.lstrip()):

                    list_words = line.split()
                    orbitals_exp[subshell] += [float(list_words[1])]
                    orbitals_basis[subshell] += [list_words[0]]

                    for x in list_of_orbitals:
                        orbitals_coeff[x] += [float(list_words[_get_column(x)])]
                    line = f.readline()
            else:
                line = f.readline()

    data = {'configuration': configuration,
            "energy": energy,
            'kinetic_energy': kinetic_energy,
            'potential_energy': potential_energy,
            'orbitals': orbitals,
            'orbitals_energy': np.array(orbitals_energy)[:, None],
            'orbitals_cusp': np.array(orbitals_cusp)[:, None],
            'orbitals_basis': orbitals_basis,
            'orbitals_exp':
                {key: np.asarray(value).reshape(len(value), 1)
                 for key, value in orbitals_exp.items()
                 if value},
            'orbitals_coeff':
                {key: np.asarray(value).reshape(len(value), 1)
                 for key, value in orbitals_coeff.items() if value != []},
            'orbitals_occupation': np.array([_get_number_of_electrons_per_orbital(configuration)[k]
                                             for k in orbitals])[:, None],
            'basis_numbers':
                {key: np.asarray([[int(x[0])] for x in value])
                 for key, value in orbitals_basis.items() if len(value) != 0}
            }

    return data



r"""
This files generates the results for Table (1) and (2) in the BFit paper.

Specifically, it optimizes the Kullback-Leibler Divergence using the fixed
 point iteration method. A normalized Gaussian model is used whose initial
 guess is the universal Gaussian basis-set multipled by two.  The constraint
 that the integral of the model should equal the atomic number is added, via
 the attribute `integral_dens`. The attribute `disp` displays the results
 at each iteration.
"""
import numpy as np

from bfit.density import SlaterAtoms
from bfit.fit import ScipyFit,KLDivergenceFPI
from bfit.grid import ClenshawRadialGrid
from bfit.model import AtomicGaussianDensity
from bfit.measure import KLDivergence
from bfit.parse_ugbs import get_ugbs_exponents
from atomdb import load

results_final = {}
atoms = ["H",  "C", "N", "O", "F", "P", "S", "Cl"]
atomic_numbs = [1, 6, 7, 8, 9, 15, 16, 17] #[1 + i for i in range(0, len(atoms))]
mult = [2, 3, 4, 3, 2, 4, 3, 2]
for k, element in enumerate(atoms):
    print("Start Atom %s" % element)

    # Construct a integration grid
    atomic_numb = atomic_numbs[k]
    grid = ClenshawRadialGrid(
        atomic_numb, num_core_pts=10000, num_diffuse_pts=899, include_origin=True#, extra_pts=[50,75,100],
    )

    # Initial Guess constructed from UGBS
    ugbs = get_ugbs_exponents(element)
    exps_s = ugbs["S"]
    num_s = len(exps_s)
    exps_p = ugbs["P"]
    num_p = len(exps_p)
    coeffs = np.array([atomic_numb / (num_s + num_p)] * (num_s + num_p))
    e_0 = np.array(exps_s + exps_p) * 2.0

    # Construct Atomic Density and Fitting Object
    #density = SlaterAtoms(element=element).atomic_density(grid.points)
    # Use AtomDB to calculate the density
    atom = load(elem=element, charge=0, mult=mult[k], dataset="hci", datapath="/home/ali-tehrani/SoftwareProjects/AtomDBdata")

    dens = atom.dens_func()
    density = dens(grid.points)

    import matplotlib.pyplot as plt
    # plt.plot(grid.points, density, "bo-")
    # plt.show()
    print(density[-1], grid.points[-1], grid.points[0:3])
    density[density < 0.0] = 0.0
    # continue
    # assert 1 == 0
    model = AtomicGaussianDensity(grid.points, num_s=num_s, num_p=num_p, normalize=True)
    fit = KLDivergenceFPI(grid, density, model, mask_value=1e-18, spherical=True,
                          integral_dens=atomic_numb)

    # Run the Kullback-Leibler FPI Method
    results = fit.run(
        coeffs, e_0, maxiter=10000, c_threshold=1e-6, e_threshold=1e-6, d_threshold=1e-14,
        disp=True
    )
    # Construct Fitting Object using SLSQP and optimizing KL
    # measure = KLDivergence(mask_value=1e-18)
    # fit_KL_slsqp = ScipyFit(grid, density, model, measure=measure, method="SLSQP", spherical=True)
    # # Run the SLSQP optimization algorithm
    # results = fit_KL_slsqp.run(coeffs, e_0, maxiter=10000, disp=True, with_constraint=True, tol=1e-14)

    print("KL-FPI INFO")
    print("-----------")
    print("Success %s" % results["success"])
    print("Final Coeffs")
    print(results["coeffs"])
    print("Final Exponents")
    print(results["exps"])
    print("Integration Value & L1 & L_infinity & LS & KL (With 4 pi r^2 included)")
    p = results["performance"][-1]
    print(p)
    # Calculate the relative errors
    spherical = 4.0 * np.pi * grid.points**2.0
    l1 = results["performance"][-1][1] / grid.integrate(density * spherical)
    linf = results["performance"][-1][2] / np.max(density)
    ls = results["performance"][-1][3] / grid.integrate(density**2.0 * spherical)
    kl = results["performance"][-1][4] / grid.integrate(
        density * np.log(density) * spherical
    )
    print("Relative Errors L1 & L_infinity & LS & KL (With 4 pi r^2 included)")
    print([l1, linf, ls, kl])

    # Store the results
    results_final[element + "_coeffs_s"] = results["coeffs"][:num_s]
    results_final[element + "_exps_s"] = results["exps"][:num_s]
    results_final[element + "_coeffs_p"] = results["coeffs"][num_s:]
    results_final[element + "_exps_p"] = results["exps"][num_s:]
    results_final[element + "_errors_4pir2"] = results["performance"][-1]
    results_final[element + "_sucess"] = results["success"]
    results_final[element + "_errors"] = [l1, linf, ls, kl]

    np.savez("./result_kl_fpi_method_cugbasis_atomdb_hci.npz", **results_final)



# -*- coding: utf-8 -*-
# BFit is a Python library for fitting a convex sum of Gaussian
# functions to any probability distribution
#
# Copyright (C) 2020- The QC-Devs Community
#
# This file is part of BFit.
#
# BFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# BFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# ---
r"""Slater Atomic Density Module."""

from bfit._slater import load_slater_wfn
import numpy as np
from scipy.special import factorial


__all__ = ["SlaterAtoms"]


class SlaterAtoms:
    r"""
    Atomic Density Class.

    Reads and parses information from the .slater file [1], [2] of an atom and stores it inside
    this class. Elements supported by default from "./bfit/data/examples/" range from Hydrogen
    to Xenon.

    Each electron of the atom is associated to a molecular spin-orbital written as:

    .. math::
        \phi_i(r, \theta, \phi) = \bigg[\sum_{j=1}^{M_i} c^i_j R_{n^i_j}(r, \alpha^i_j) \bigg]
        Y_{l_i}^{m_i}(\theta, \phi) \sigma(m_i),

    where :math:`R_{n^i_j}(r)` is a Slater-type orbital with quantum number :math:`n^i_j`,
    :math:`c^i_j` is the coefficient of the jth Slater-type orbital,
    :math:`\alpha_j^i` is the exponent of the jth Slater-type orbital,
    :math:`Y_{l_i}^{m_i}` is the spherical harmonic with complex form, angular momentum :math:`l_i`
    determined by the electron in the electron configuration and :math:`m_i` is the spin determined
    by applying Hund's rule to the electron configuration, and
    :math:`\sigma` is the spin-function determined from applying Hund's rule to the electron.

    References
    ----------
    .. [1] Koga, T. , Kanayama, K. , Watanabe, S. and Thakkar, A. J. (1999),
       Analytical Hartree–Fock wave functions subject to cusp and asymptotic
       constraints: He to Xe, Li+ to Cs+, H−  Int. J. Quantum Chem., 71: 491-497.
       doi:10.1002/(SICI)1097-461X(1999)71:6<491::AID-QUA6>3.0.CO;2-T

    .. [2] Koga, T., Kanayama, K., Watanabe, T. et al. Analytical Hartree–Fock wave functions
       for the atoms Cs to Lr. Theor Chem Acc 104, 411–413 (2000).
       https://doi.org/10.1007/s002140000150

    """

    def __init__(self, element, anion=False, cation=False):
        r"""
        Construct SlaterAtoms object.

        Parameters
        ----------
        element : str
            Symbol of element.
        anion : bool
            If true, then the anion of element is used. Some elements do not have
            anion information.
        cation : bool
            If true, then the cation of element is used. Some elements do not
            have cation information.

        """
        if not isinstance(element, str) or not element.isalpha():
            raise TypeError("The element argument should be all letters string.")
        if anion and cation:
            raise ValueError("Both anion and cation cannot be true.")

        data = load_slater_wfn(element, anion, cation)
        self._energy = data["energy"]
        self._potential_energy = data["potential_energy"]
        self._kinetic_energy = data["kinetic_energy"]
        self._configuration = data["configuration"]
        self._orbitals = data["orbitals"]
        self._orbitals_occupation = data["orbitals_occupation"]
        self._orbitals_basis = data["orbitals_basis"]
        self._basis_numbers = data["basis_numbers"]
        self._orbitals_exp = data["orbitals_exp"]
        self._orbitals_coeff = data["orbitals_coeff"]
        self._orbitals_energy = data["orbitals_energy"]
        self._orbitals_cusp = data["orbitals_cusp"]

    @property
    def energy(self):
        r"""Energy of atom."""
        return self._energy

    @property
    def kinetic_energy(self):
        r"""Kinetic energy of atom."""
        return self._kinetic_energy

    @property
    def potential_energy(self):
        r"""Potential energy of atom."""
        return self._potential_energy

    @property
    def configuration(self):
        r"""Return string representing the electron configuration of the atom.

        The electron configuration of the atom is written in form that writes out the
        atomic subshells with the number of electrons assigned to that atomic subshell.
        For example, Beryllium returns "1S(2)2S(2)".
        """
        return self._configuration

    @property
    def orbitals(self):
        r"""
        List of strings representing each of the atomic subshells in the electron configuration.

        For example, Beryllium returns ["1S", "2S"] in its electron configuration.
        Ordered based on "S", "P", "D", etc.
        """
        return self._orbitals

    @property
    def orbitals_occupation(self):
        r"""
        Array returning number of electrons in each of the orbitals in the electron configuration.

        For example, Beryllium returns ndarray([[2], [2]]).
        """
        return self._orbitals_occupation

    @property
    def orbitals_basis(self):
        r"""
        Return grouping of Slater-type orbital to the azimuthal quantum number ("S", "P", ...).

        Dictionary mapping type of orbital (e.g. "S", "P") to the number
        and type of the :math:`N` Slater-type orbital. For example, Helium would
        map "S" to ['2S', '1S', '1S', '1S', '2S']. This implies that all
        molecular orbitals corresponding to s-orbital will have it's radial component
        expanded in that Slater-type orbitals according to the label.
        """
        return self._orbitals_basis

    @property
    def basis_numbers(self):
        r"""
        Return type of Slater-type orbital to the type, e.g. "S".

        Dictionary mapping type of orbital (e.g. "S", "P") to array
        containing :math:`n` of the :math:`N` Slater-type orbitals. These play the
        role of the principal quantum number to each Slater-type orbitals.
        With the Helium example, "S" will map to [[2], [1], [1], [1], [2]].
        """
        return self._basis_numbers

    @property
    def orbitals_exp(self):
        r"""
        Exponent of each Slater-type orbital grouped by type of orbital.

        Dictionary mapping type of orbitals (e.g. "S", "P") to the
        exponent :math:`\alpha_j^i` of each of the :math:`M_i` Slater-type orbitals.
        """
        return self._orbitals_exp

    @property
    def orbitals_coeff(self):
        r"""
        Coefficients of each Slater-type orbital grouped by type of orbital.

        Dictionary mapping the molecular orbital (e.g. "1S", "2S", ..) to
        the coefficients :math:`c^i_j` of expansion w.r.t. the :math:`M_i` Slater-type orbital.
        """
        return self._orbitals_coeff

    @property
    def orbitals_energy(self):
        r"""Energy of each of the :math:`N` Slater-type orbital."""
        return self._orbitals_energy

    @property
    def orbitals_cusp(self):
        r"""
        Cusp values of each of the N Slater-type orbital.

        Same ordering as `orbitals`. Does not exist for Heavy atoms past Xenon.
        """
        return self._orbitals_cusp

    @staticmethod
    def radial_slater_orbital(exponent, number, points):
        r"""
        Compute the radial component of Slater-type orbitals on the given points.

        The radial component of the Slater-type orbital is defined as:

        .. math::
            R(r) = N r^{n-1} e^{- \alpha r}

        where,
        :math:`n` is the principal quantum number of that orbital,
        :math:`N` is the normalizing constant,
        :math:`r` is the radial point, distance to the origin, and
        :math:`\alpha` is the zeta exponent of that orbital.

        Parameters
        ----------
        exponent : ndarray, (M, 1)
            The zeta exponents :math:`\zeta` of :math:`M` Slater orbitals.
        number : ndarray, (M, 1)
            The principal quantum numbers :math:`n` of :math:`M` Slater orbitals.
        points : ndarray, (N,)
            The radial :math:`r` grid points.

        Returns
        -------
        slater : ndarray, (N, M)
            The :math:`M` Slater-type orbitals evaluated on :math:`N` grid points.

        Notes
        -----
        - The principal quantum number of all functions are stored in `basis_numbers`.

        - The alpha exponents of all functions are stored in the attribute `orbitals_exp`.

        """
        if points.ndim != 1:
            raise ValueError("The argument point should be a 1D array.")
        # compute norm & pre-factor
        norm = np.power(2. * exponent, number) * np.sqrt((2. * exponent) / factorial(2. * number))
        pref = np.power(points, number - 1).T
        # compute slater orbitals
        slater = norm.T * pref * np.exp(-exponent * points).T
        return slater

    def phi_matrix(self, points, deriv=False):
        r"""
        Compute the linear combination of Slater-type orbitals on the given points.

        Each row corresponds to a point on the grid, represented as :math:`r` and
        each column is represented as a linear combination of Slater-type atomic orbitals
        of the form:

        .. math::
            \sum_{i=1}^{M} c_i R(r, n_i, \alpha_i)

        where,
        :math:`c_i` is the coefficient of the Slater-type orbital,
        :math:`\alpha_i` is the zeta exponent attached to the Slater-type orbital,
        :math:`n_i` is the principal quantum number attached to the Slater-type orbital,
        :math:`R(r, n_i, C_i)` is the radial component of the Slater-type orbital,
        :math:`M` is the number of orbitals.

        Parameters
        ----------
        points : ndarray, (N,)
            The radial grid points.
        deriv : bool
            If true, use the derivative of the slater-orbitals.

        Returns
        -------
        phi_matrix : ndarray(N, K)
            The linear combination of Slater-type orbitals evaluated on the :math:`N` grid points,
            and :math:`M` is the number of atomic subshells (ignoring spin) within the electron
            configuration. The order is S orbitals, then P then D and spin is ignored.

        Notes
        -----
        - At r = 0, the derivative of slater-orbital is undefined and this function returns
          zero instead. See "derivative_radial_slater_type_orbital".

        """
        # compute orbital composed of a linear combination of Slater
        phi_matrix = np.zeros((len(points), len(self.orbitals)))
        for index, orbital in enumerate(self.orbitals):
            exps, number = self.orbitals_exp[orbital[1]], self.basis_numbers[orbital[1]]
            if deriv:
                slater = self.derivative_radial_slater_type_orbital(exps, number, points)
            else:
                slater = self.radial_slater_orbital(exps, number, points)
            phi_matrix[:, index] = np.dot(slater, self.orbitals_coeff[orbital]).ravel()
        return phi_matrix

    def atomic_density(self, points, mode="total"):
        r"""
        Compute atomic density on the given points.

        The total density is written as a linear combination of molecular orbitals squared
        whose coefficients is the orbital occupation number of the electron configuration:

        .. math::
            \sum n_i |\phi_i(r)|^2

        where,
        :math:`n_i` is the number of electrons in the ith molecular orbital,
        :math:`\phi_i(r)` is the ith molecular orbital, whose radial component is
        a linear combination of Slater-type orbitals evaluated on the point :math:`r` and
        whose angles within the spherical coordinates are integrated.

        For core and valence density, please see more Info below.

        Parameters
        ----------
        points : ndarray, (N,)
            The radial grid points.
        mode : str
            The type of atomic density, which can be "total", "valence" or "core".

        Returns
        -------
        dens : ndarray, (N,)
            The atomic density on the grid points.

        Notes
        -----
        The core density and valence density is respectively written as:

        .. math::
            \begin{align*}
                \rho^{core}(r) &= \sum n_i (1 - e^{-|e_i - e_{homo}|^2}) |\phi_i(r)| \\
                \rho^{valence}(r) &= \sum n_i e^{-|e_i - e_{homo}|^2} |\phi_i(r)|
            \end{align*}

        where,
        :math:`e_i` is the energy of the orbital i.
        :math:`e_{HOMO}` is the energy of the highest occupying orbital.

        """
        if mode not in ["total", "valence", "core"]:
            raise ValueError("Argument mode not recognized!")

        # compute orbital occupation numbers
        orb_occs = self.orbitals_occupation
        if mode == "valence":
            orb_homo = self.orbitals_energy[len(self.orbitals_occupation) - 1]
            orb_occs = orb_occs * np.exp(-(self.orbitals_energy - orb_homo)**2)
        elif mode == "core":
            orb_homo = self.orbitals_energy[len(self.orbitals_occupation) - 1]
            orb_occs = orb_occs * (1. - np.exp(-(self.orbitals_energy - orb_homo)**2))
        # compute density
        dens = np.dot(self.phi_matrix(points)**2, orb_occs).ravel() / (4 * np.pi)
        return dens

    @staticmethod
    def derivative_radial_slater_type_orbital(exponent, number, points):
        r"""
        Compute the derivative of the radial component of Slater-type orbital on the given points.

        The derivative of the Slater-type orbital is defined as:

        .. math::
            \frac{d R(r)}{dr} = \bigg(\frac{n-1}{r} - \alpha \bigg) N r^{n-1} e^{- \alpha r},

        where
        :math:`n` is the principal quantum number of that orbital,
        :math:`N` is the normalizing constant,
        :math:`r` is the radial point, distance to the origin, and
        :math:`\alpha` is the zeta exponent of that orbital.

        Parameters
        ----------
        exponent : ndarray, (M, 1)
            The zeta exponents of Slater orbitals.
        number : ndarray, (M, 1)
            The principle quantum numbers of Slater orbitals.
        points : ndarray, (N,)
            The radial grid points. If points contain zero, then it is undefined at those
            points and set to zero.

        Returns
        -------
        slater : ndarray, (N, M)
            The Slater-type orbitals evaluated on the :math:`N` grid points.

        Notes
        -----
        - At r = 0, the derivative is undefined and this function returns zero instead.

        References
        ----------
        See wikipedia page on "Slater-Type orbitals".

        """
        slater = SlaterAtoms.radial_slater_orbital(exponent, number, points)
        # Consider the case when dividing by zero.
        with np.errstate(divide='ignore'):
            # derivative
            deriv_pref = (number.T - 1.) / np.reshape(points, (points.shape[0], 1)) - exponent.T
            deriv_pref[np.abs(points) < 1e-10, :] = 0.0
        deriv = deriv_pref * slater
        return deriv

    def positive_definite_kinetic_energy(self, points):
        r"""
        Positive definite or Lagrangian kinetic energy density.

        Parameters
        ----------
        points : ndarray,(N,)
            The radial grid points.

        Returns
        -------
        energy : ndarray, (N,)
            The kinetic energy on the grid points.

        """
        phi_matrix = np.zeros((len(points), len(self.orbitals)))
        angular = {
            "S": 0.0, "P": 2.0, "D": 6.0, "F": 12.0
        }
        for index, orbital in enumerate(self.orbitals):
            exps, numbers = self.orbitals_exp[orbital[1]], self.basis_numbers[orbital[1]]
            # Calculate del^2 of radial component
            # derivative of the radial component
            deriv_radial = self.derivative_radial_slater_type_orbital(exps, numbers, points)
            phi_matrix[:, index] = np.ravel(np.dot(deriv_radial, self.orbitals_coeff[orbital])**2.0)

            # Calculate del^2 of spherical component
            slater = SlaterAtoms.radial_slater_orbital(exps, numbers, points)
            with np.errstate(divide='ignore'):
                gradient_sph = (
                    angular[orbital[1]] *
                    np.ravel(np.dot(slater, self.orbitals_coeff[orbital]))**2.0 /
                    points**2.0
                )
                gradient_sph[np.abs(points) < 1e-10] = 0.0
            phi_matrix[:, index] += gradient_sph

        orb_occs = self.orbitals_occupation
        energy = np.dot(phi_matrix, orb_occs).ravel() / 2.
        return energy / (4.0 * np.pi)

    def derivative_density(self, points):
        r"""
        Return the derivative of the atomic density on a set of points.

        Parameters
        ----------
        points : ndarray(N,)
            The :math:`N` radial grid points.

        Returns
        -------
        deriv : ndarray(N,)
            The derivative of atomic density on the grid points.

        """
        factor = self.phi_matrix(points) * self.phi_matrix(points, deriv=True)
        derivative = np.dot(2. * factor, self.orbitals_occupation).ravel() / (4 * np.pi)
        return derivative



r"""
This files generates the results for Table (1) and (2) in the BFit paper.

Specifically, it optimizes the Kullback-Leibler Divergence using the fixed
 point iteration method. A normalized Gaussian model is used whose initial
 guess is the universal Gaussian basis-set multipled by two.  The constraint
 that the integral of the model should equal the atomic number is added, via
 the attribute `integral_dens`. The attribute `disp` displays the results
 at each iteration.
"""
import numpy as np

from bfit.density import SlaterAtoms
from bfit.fit import KLDivergenceFPI
from bfit.grid import ClenshawRadialGrid
from bfit.model import AtomicGaussianDensity
from bfit.parse_ugbs import get_ugbs_exponents


results_final = {}
atoms = ["h", "he", "li", "be", "b", "c", "n", "o", "f", "ne",
         "na", "mg", "al", "si", "p", "s", "cl", "ar", "k", "ca",
         "sc", "ti", "v", "cr", "mn", "fe", "co", "ni", "cu",
         "zn", "ga", "ge", "as", "se", "br", "kr", "rb", "sr",
         "y", "zr", "nb", "mo", "tc", "ru", "rh", "pd", "ag",
         "cd", "in", "sn", "sb", "te", "i", "xe"]
atomic_numbs = [1 + i for i in range(0, len(atoms))]

for k, element in enumerate(atoms):
    print("Start Atom %s" % element)

    # Construct a integration grid
    atomic_numb = atomic_numbs[k]
    grid = ClenshawRadialGrid(
        atomic_numb, num_core_pts=10000, num_diffuse_pts=899, extra_pts=[50, 75, 100]
    )

    # Initial Guess constructed from UGBS
    ugbs = get_ugbs_exponents(element)
    exps_s = ugbs["S"]
    num_s = len(exps_s)
    exps_p = ugbs["P"]
    num_p = len(exps_p)
    coeffs = np.array([atomic_numb / (num_s + num_p)] * (num_s + num_p))
    e_0 = np.array(exps_s + exps_p) * 2.0

    # Construct Atomic Density and Fitting Object
    density = SlaterAtoms(element=element).atomic_density(grid.points)
    model = AtomicGaussianDensity(grid.points, num_s=num_s, num_p=num_p, normalize=True)
    fit = KLDivergenceFPI(grid, density, model, mask_value=1e-18, spherical=True,
                          integral_dens=atomic_numb)

    # Run the Kullback-Leibler FPI Method
    results = fit.run(
        coeffs, e_0, maxiter=10000, c_threshold=1e-6, e_threshold=1e-6, d_threshold=1e-14,
        disp=True
    )

    print("KL-FPI INFO")
    print("-----------")
    print("Success %s" % results["success"])
    print("Final Coeffs")
    print(results["coeffs"])
    print("Final Exponents")
    print(results["exps"])
    print("Integration Value & L1 & L_infinity & LS & KL (With 4 pi r^2 included)")
    p = results["performance"][-1]
    print(p)
    # Calculate the relative errors
    spherical = 4.0 * np.pi * grid.points**2.0
    l1 = results["performance"][-1][1] / grid.integrate(density * spherical)
    linf = results["performance"][-1][2] / np.max(density)
    ls = results["performance"][-1][3] / grid.integrate(density**2.0 * spherical)
    kl = results["performance"][-1][4] / grid.integrate(
        density * np.log(density) * spherical
    )
    print("Relative Errors L1 & L_infinity & LS & KL (With 4 pi r^2 included)")
    print([l1, linf, ls, kl])

    # Store the results
    results_final[element + "_coeffs"] = results["coeffs"]
    results_final[element + "_exps"] = results["exps"]

np.save("./result_kl_fpi_method.npz", results_final)



# -*- coding: utf-8 -*-
# BFit is a Python library for fitting a convex sum of Gaussian
# functions to any probability distribution
#
# Copyright (C) 2020- The QC-Devs Community
#
# This file is part of BFit.
#
# BFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# BFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# ---
r"""Grid module for integration."""

import numpy as np

__all__ = ["ClenshawRadialGrid", "UniformRadialGrid", "CubicGrid"]


class _BaseRadialGrid:
    r"""Radial Grid Base Class."""

    def __init__(self, points):
        """
        Construct BaseRadialGrid object.

        Parameters
        ----------
        points : ndarray(N,)
            The radial grid points.

        """
        if not isinstance(points, np.ndarray) or points.ndim != 1:
            raise TypeError("Argument points should be a 1D numpy array.")
        self._points = np.ravel(points)

    @property
    def points(self):
        """Radial grid points."""
        return self._points

    def __len__(self):
        """Return number of grid points."""
        return self._points.shape[0]

    def integrate(self, arr):
        r"""
        Compute trapezoidal integration of a function evaluated on the radial grid points.

        In other words, :math:`\int f(r) dr`, where :math:'f(r)' is the integrand.

        Parameters
        ----------
        arr : ndarray(N,)
            The integrand evaluated on the :math:`N` radial grid points.

        Returns
        -------
        float :
            The value of the integral.

        """
        if arr.shape != self.points.shape:
            raise ValueError(
                f"The argument arr should have {self.points.shape} shape!"
            )
        return np.trapz(y=arr, x=self.points)


class UniformRadialGrid(_BaseRadialGrid):
    r"""
    Uniformly Distributed Radial Grid Class.

    The grid points are equally-spaced in :math:`[0, K)` interval, where K is the upper bound
    on the grid.
    """

    def __init__(self, num_pts, min_radii=0., max_radii=100., dtype=np.longdouble):
        """
        Construct the UniformRadialGrid object.

        Parameters
        ----------
        num_pts : int
            The number of grid points.
        min_radii : float, optional
            The smallest radial grid point.
        max_radii : float, optional
            The largest radial grid point.
        dtype : data-type, optional
            The desired NumPy data-type.

        """
        if not isinstance(num_pts, int) or num_pts <= 0:
            raise TypeError(f"Argument num_pts {type(num_pts)} should be an integer.")
        if not isinstance(min_radii, (int, float)):
            raise TypeError(
                f"Argument min_radii {type(min_radii)} should be a float or integer."
            )
        if not isinstance(max_radii, (int, float)):
            raise TypeError(
                f"Argument max_radii {type(max_radii)} should be a float or integer."
            )
        if max_radii <= min_radii:
            raise ValueError("The max_radii should be greater than the min_radii.")

        # compute points
        points = np.linspace(start=min_radii, stop=max_radii, num=num_pts, dtype=dtype)
        super().__init__(points)


class ClenshawRadialGrid(_BaseRadialGrid):
    r"""Clenshaw-Curtis Radial Grid Class.

    The Clenshaw-Curtis grid places more points closer to the origin of the interval
    :math:`[0, \inf).`
    It is defined as follows. Let :math:`Z, m, n` be the atomic number, number of points near
    the origin, and the number of points far from the origin, respectively.

    Then each point :math:`r_p` of the Clenshaw radial grid is either

    .. math::
        \begin{eqnarray}
            r_p = \frac{1}{2Z} \bigg(1 - \cos\bigg(\frac{\pi p}{400} \bigg)\bigg)  &
            p = 0, 1, \cdots, m - 1 \\
            r_p = 25 \bigg(1 - \cos\bigg(\frac{\pi p}{600} \bigg)\bigg) & p = 1, \cdots, n - 1\\
        \end{eqnarray}
    """

    def __init__(self, atomic_number, num_core_pts, num_diffuse_pts, extra_pts=None,
                 include_origin=True, dtype=np.longdouble):
        r"""
        Construct ClenshawRadialGrid grid object.

        Parameters
        ----------
        atomic_number : int
            The atomic number of the atom for which the grid is generated.
        num_core_pts : int
            The number of points near the origin/core region.
        num_diffuse_pts : int
            The number of points far from the origin/core region.
        extra_pts : list
            Additional points to be added to the grid, commonly points far away from origin.
        include_origin : bool
            If true, then include the origin :math:`r=0`.
        dtype : data-type, optional
            The desired NumPy data-type.

        """
        if not isinstance(atomic_number, int) or atomic_number <= 0:
            raise TypeError("Argument atomic_number should be a positive integer.")
        if not isinstance(num_core_pts, int) or num_core_pts < 0:
            raise TypeError("Argument numb_core_pts should be a non-negative integer.")
        if not isinstance(num_diffuse_pts, int) or num_diffuse_pts < 0:
            raise TypeError("Argument num_diffuse_pts should be a non-negative integer.")
        if not isinstance(include_origin, bool):
            raise TypeError(
                f"Argument include_origin {type(include_origin)} should be of type boolean."
            )

        self._atomic_number = atomic_number

        # compute core and diffuse points
        core_points = self._get_points(
            num_core_pts, mode="core", include_origin=include_origin, dtype=dtype
        )
        diff_points = self._get_points(num_diffuse_pts, mode="diffuse", dtype=dtype)

        # put all points together (0.0 is also contained in diff_points, so it should be removed)
        if extra_pts:
            # check extra points
            if not hasattr(extra_pts, '__iter__') or isinstance(extra_pts, str):
                raise TypeError("Argument extra_pts should be an iterable.")
            points = np.concatenate((core_points, diff_points, extra_pts))
        else:
            points = np.concatenate((core_points, diff_points))

        super().__init__(np.sort(points))

    @property
    def atomic_number(self):
        """Return the atomic number."""
        return self._atomic_number

    def _get_points(self, num_pts, mode="core", include_origin=True, dtype=np.longdouble):
        r"""Generate radial points on :math:`[0, \inf)` based on Clenshaw-Curtis grid.

        The "core" points are concentrated near the origin based on:

        .. math::
            r_p = 25 \bigg(1 - \cos\bigg(\frac{\pi p}{2N})\bigg) \quad \text{for } p =0,1,
            \cdots, N-1

        The "diffuse" points are concentrated away from the origin based on:

        .. math::
            r_p = \frac{1}{2Z} \bigg(1 - \cos\bigg(\frac{p\pi}{2N}\bigg)\bigg) \quad \text{for }
            p=0,1,\cdots, N-1,

        where :math:`Z` is the atomic number and :math:`N` is the number of points.

        Parameters
        ----------
        num_pts : int
            The number of points.
        mode : str, optional
            If "core", the points are placed closer to the origin. If "diffuse", the points are
            placed far away from origin.
        include_origin : bool
            If true, then include the origin when `mode`="core".
        dtype : data-type, optional
            The desired NumPy data-type.

        Returns
        -------
        points : ndarray, (N,)
            The 1D array of grid points.

        """
        if mode.lower() == "core":
            points = 1. - np.cos(
                0.5 * np.pi * np.arange(
                    (1.0 - float(include_origin)), num_pts + int(not include_origin), dtype=dtype
                ) / num_pts
            )
            points /= 2 * self._atomic_number
        elif mode.lower() == "diffuse":
            points = 25. * (
                    1. - np.cos(0.5 * np.pi * np.arange(1.0, num_pts + 1, dtype=dtype) / num_pts)
            )
        else:
            raise ValueError(
                f"Arguments mode={mode.lower()} is not recognized!"
            )
        return points


class CubicGrid:
    r"""Equally-Spaced 3D Cubic Grid Class."""

    def __init__(self, origin, axes, shape):
        """
        Construct the CubicGrid object.

        Parameters
        ----------
        origin : float
            The origin (left-most, down-most) of the 3D cubic grid.
        axes : ndarray(3, 3)
            The axes that point to the direction of the grid.
        shape : (int, int, int)
            The number of points in each axes.

        """
        # TODO: Add raise error for Type and Values here for origin, axes.
        self._axes = axes
        self._origin = origin
        dim = self._origin.size
        # Make an array to store coordinates of grid points
        self._points = np.zeros((np.prod(shape), dim))
        coords = np.array(
            np.meshgrid(
                np.arange(shape[0]), np.arange(shape[1]), np.arange(shape[2])
            )
        )
        coords = np.swapaxes(coords, 1, 2)
        coords = coords.reshape(3, -1)
        points = coords.T.dot(self._axes) + origin
        # assign the weights
        weights = self._choose_weight_scheme(shape)
        self._weights = weights
        self._points = np.array(points)

    @property
    def axes(self):
        r"""Return the axes/three-directions of the cubic grid."""
        return self._axes

    @classmethod
    def from_molecule(
            cls,
            atcorenums,
            atcoords,
            spacing=0.2,
            extension=5.0,
            rotate=True,
    ):
        r"""
        Construct a uniform grid given the molecular pseudo-numbers and coordinates.

        Parameters
        ----------
        atcorenums : np.ndarray, shape (M,)
            Pseudo-number of :math:`M` atoms in the molecule.
        atcoords : np.ndarray, shape (M, 3)
            Cartesian coordinates of :math:`M` atoms in the molecule.
        spacing : float, optional
            Increment between grid points along :math:`x`, :math:`y`, and :math:`z` direction.
        extension : float, optional
            The extension of the length of the cube on each side of the molecule.
        rotate : bool, optional
            When True, the molecule is rotated so the axes of the cube file are
            aligned with the principle axes of rotation of the molecule.
            If False, generates axes based on the x,y,z-axis and the spacing parameter, and
            the origin is defined by the maximum/minimum of the atomic coordinates.
        """
        # calculate center of mass of the nuclear charges:
        totz = np.sum(atcorenums)
        com = np.dot(atcorenums, atcoords) / totz
        # Determine best axes and coordinates to calculate the lower and upper bound of grid.
        if rotate:
            # calculate moment of inertia tensor:
            itensor = np.zeros([3, 3])
            for i in range(atcorenums.shape[0]):
                xyz = atcoords[i] - com
                r = np.linalg.norm(xyz) ** 2.0
                tempitens = np.diag([r, r, r])
                tempitens -= np.outer(xyz.T, xyz)
                itensor += atcorenums[i] * tempitens
            # Eigenvectors define the new axes of the grid with spacing
            _, v = np.linalg.eigh(itensor)
            # Project the coordinates of atoms centered at the center of mass to the eigenvectors
            new_coordinates = np.dot((atcoords - com), v)
            axes = spacing * v
        else:
            # Just use the original coordinates
            new_coordinates = atcoords
            # Compute the unit vectors of the cubic grid's coordinate system
            axes = np.diag([spacing, spacing, spacing])

        # maximum and minimum value of x, y and z coordinates/grid.
        max_coordinate = np.amax(new_coordinates, axis=0)
        min_coordinate = np.amin(new_coordinates, axis=0)
        # Compute the required number of points along x, y, and z axis
        shape = (max_coordinate - min_coordinate + 2.0 * extension) / spacing
        # Add one to include the upper-bound as well.
        shape = np.ceil(shape)
        shape = np.array(shape, int)
        # Compute origin by taking the center of mass then subtracting the half of the number
        #    of points in the direction of the axes.
        origin = com - np.dot((0.5 * shape), axes)
        return cls(origin, axes, shape)

    def _calculate_volume(self, shape):
        r"""Return the volume of the Uniform Grid."""
        # Shape needs to be an argument, because I need to calculate the weights before
        #       initializing the _HyperRectangleGrid (where shape is set there).
        # Three-Dims: Volume of a parallelepiped spanned by a, b, c is  | (a x b) dot c|.
        if len(shape) == 3:
            volume = np.dot(
                np.cross(shape[0] * self.axes[0], shape[1] * self.axes[1]),
                shape[2] * self.axes[2],
            )
        else:
            # Two-Dims: Volume of a parallelogram is the absolute value of the determinant |a x b|
            volume = np.linalg.det(
                np.array([shape[0] * self.axes[0], shape[1] * self.axes[1]])
            )
        return np.abs(volume)

    def _choose_weight_scheme(self, shape):
        # Choose different weighting schemes.
        volume = self._calculate_volume(shape)
        numpnt = 1.0 * np.prod(shape)
        weights = np.full(np.prod(shape), volume / numpnt)
        return weights

    @property
    def points(self):
        """Return cubic grid points."""
        return self._points

    def __len__(self):
        """Return the number of grid points."""
        return self._points.shape[0]

    def integrate(self, arr):
        r"""Compute the integral of a function evaluated on the grid points based on Riemann sums.

        .. math::
            \int\int\int f(x, y, z) dx dy dz \approx \sum_i \sum_j \sum_k f(x_i, y_j, z_k) w_{ijk}

        where :math:'f(r)' is the integrand, and :math:`w_{ijk}` is the weight associated with
        the (i, j, k)th point.

        Parameters
        ----------
        arr : ndarray
            The integrand evaluated on the grid points.

        Returns
        -------
        value : float
            The value of the integral.

        """
        if arr.shape != (len(self),):
            raise ValueError(
                f"Argument arr should have ({len(self)},) shape."
            )
        value = np.sum(self._weights * arr)
        return value



# -*- coding: utf-8 -*-
# BFit is a Python library for fitting a convex sum of Gaussian
# functions to any probability distribution
#
# Copyright (C) 2020- The QC-Devs Community
#
# This file is part of BFit.
#
# BFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# BFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# ---
"""Measure Module."""

from abc import ABC, abstractmethod

import numpy as np


__all__ = ["SquaredDifference", "KLDivergence", "TsallisDivergence"]


class Measure(ABC):
    r"""Abstract base class for the measures."""

    @abstractmethod
    def evaluate(self, density, model, deriv=False):
        r"""
        Abstract method for evaluating the measure.

        Parameters
        ----------
        density : ndarray(N,)
            The exact density evaluated on the grid points.
        model : ndarray(N,)
            The model evaluated on the same :math:`N` points that `density` is
            evaluated on.
        deriv : bool, optional
            Whether it should return the derivatives of the measure w.r.t. to the
            model parameters.

        Returns
        -------
        m : ndarray(N,)
            The measure between density & model on the grid points.
        dm : ndarray(N,)
            The derivative of measure w.r.t. model evaluated on the
            grid points. Only returned if `deriv=True`.

        """
        raise NotImplementedError("Evaluate function should be implemented.")


class SquaredDifference(Measure):
    r"""Squared Difference Class for performing the Least-Squared method."""

    def evaluate(self, density, model, deriv=False):
        r"""
        Evaluate squared difference b/w density & model on the grid points.

        This is defined to be :math:`(f(x) - g(x))^2`,
        where,
        :math:`f` is the true function, and
        :math:`g` is the model function,

        Parameters
        ----------
        density : ndarray(N,)
            The exact density evaluated on the grid points.
        model : ndarray(N,)
            The model density evaluated on the grid points.
        deriv : bool, optional
            Whether to compute the derivative of squared difference w.r.t. model density.
            Default is false.

        Returns
        -------
        m : ndarray(N,)
            The squared difference between density & model on the grid points.
        dm : ndarray(N,)
            The derivative of squared difference w.r.t. model density evaluated on the
            grid points, only returned if `deriv=True`.

        Notes
        -----
        - This class returns the squared difference at each point in the domain.
          One would need to integrate this to get the desired least-squares.

        """
        if not isinstance(model, np.ndarray):
            raise ValueError(
                f"Argument model should be {density.shape} array."
            )
        if not isinstance(density, np.ndarray) or density.ndim != 1:
            raise ValueError("Arguments density should be a 1D numpy array.")
        if model.shape != density.shape:
            raise ValueError(f"Model shape {model.shape} should be the same as density"
                             f" {density.shape}.")
        if not isinstance(deriv, bool):
            raise TypeError(f"Deriv {type(deriv)} should be Boolean type.")
        # compute residual
        residual = density - model
        # compute squared residual
        value = np.power(residual, 2)
        # compute derivative of squared residual w.r.t. model
        if deriv:
            return value, -2 * residual
        return value


class KLDivergence(Measure):
    r"""Kullback-Leibler Divergence Class."""

    def __init__(self, mask_value=1.e-12, negative_val=100000.0):
        r"""
        Construct the Kullback-Leibler class.

        Parameters
        ----------
        mask_value : float, optional
            The elements less than or equal to this number are masked in a division,
            and then replaced with the value of one so that logarithm of one is zero.
        negative_val : (float, np.inf), optional
            Constant value that gets returned if the model density is negative
            (i.e. not a true probability distribution). Useful for optimization algorithms
            with weak constraints.

        """
        super().__init__()
        self._mask_value = mask_value
        self._negative_val = negative_val

    @property
    def mask_value(self):
        r"""Masking value used when evaluating the measure."""
        return self._mask_value

    @property
    def negative_val(self):
        r"""Value that gets returned if the model density is negative."""
        return self._negative_val

    def evaluate(self, density, model, deriv=False):
        r"""
        Evaluate the integrand of Kullback-Leibler divergence b/w true & model.

        .. math ::
            D(f, g) := \int_G f(x) \ln \bigg( \frac{f(x)}{g(x)} \bigg) dx

        where,
        :math:`f` is the true probability distribution,
        :math:`g` is the model probability distribution,
        :math:`G` is the grid being integrated on.

        If the model density is negative, then this function will return infinity.
        If :math:`g(x) < \text{mask value}` then :math:`\frac{f(x)}{g(x)} = 1`.

        Parameters
        ----------
        density : ndarray(N,)
            The exact density evaluated on the grid points.
        model : ndarray(N,)
            The model density evaluated on the grid points.
        deriv : bool, optional
            Whether to return the derivative of divergence w.r.t. model density, as well.
            Default is false.

        Returns
        -------
        m : ndarray(N,)
            The Kullback-Leibler divergence between density & model on the grid points.
        dm : ndarray(N,)
            The derivative of divergence w.r.t. model density evaluated on the grid points.
            Only returned if `deriv=True`.

        Raises
        ------
        ValueError :
            If the model density is negative, then the integrand is un-defined.

        Notes
        -----
        - Values of nodel density that are less than `mask_value` are masked when used in
          division and then replaced with the value of 1 so that logarithm of one is zero.
        - This class does not return the Kullback-Leibler divergence.
          One would need to integrate this to get the KL value.

        """
        # check model density
        if not isinstance(model, np.ndarray):
            raise ValueError(
                f"Argument model should be {density.shape} array."
            )
        if not isinstance(density, np.ndarray) or density.ndim != 1:
            raise ValueError("Arguments density should be a 1D numpy array.")
        if model.shape != density.shape:
            raise ValueError(f"Model shape {model.shape} should be the same as density"
                             f" {density.shape}.")
        if not isinstance(deriv, bool):
            raise TypeError(f"Deriv {type(deriv)} should be Boolean type.")
        if np.any(model < 0.):
            if deriv:
                # Add an incredibly large derivative
                return np.array([self.negative_val] * model.shape[0]), \
                       np.array([self.negative_val] * model.shape[0])
            return np.array([self.negative_val] * model.shape[0])

        # compute ratio & replace masked values by 1.0
        ratio = density / np.ma.masked_less_equal(model, self.mask_value)
        ratio = np.ma.filled(ratio, fill_value=1.0)

        # compute KL divergence
        # Add ignoring division by zero and multiplying by np.nan
        with np.errstate(divide='ignore', invalid="ignore"):
            value = density * np.log(ratio)
        # compute derivative
        if deriv:
            return value, -ratio
        return value


class TsallisDivergence(Measure):
    r"""Tsallis Divergence Class."""

    def __init__(self, alpha=1.001, mask_value=1.e-12):
        r"""
        Construct the Kullback-Leibler class.

        Parameters
        ----------
        alpha : float
            The alpha parameter of the Tsallis divergence. If it tends towards
            one, then Tsallis divergence approaches Kullback-Leibler. See Notes
            for more information.
        mask_value : float, optional
            The elements less than or equal to this number are masked in a division,
            and then replaced with the value of one so that logarithm of one is zero.

        Notes
        -----
        - Care should be taken on :math:`\alpha`. If :math:`\alpha > 1` and it isn't too
          close to 1, then Tsallis divergence upper bounds the Kullback-Leibler and can be useful
          for optimization purposes.

        """
        self._alpha = alpha
        self._mask_value = mask_value
        if self._alpha < 0:
            raise ValueError(f"Alpha parameter {alpha} should be positive.")
        if np.abs(self._alpha - 1.0) < 1e-10:
            raise ValueError(f"Alpha parameter {alpha} shouldn't be equal to one."
                             f"Use Kullback-Leibler divergence instead.")
        super().__init__()

    @property
    def alpha(self):
        r"""Return alpha parameter of the Tsallis divergence."""
        return self._alpha

    @property
    def mask_value(self):
        r"""Return masking value used when evaluating the measure."""
        return self._mask_value

    def evaluate(self, density, model, deriv=False):
        r"""
        Evaluate the integrand of Tsallis divergence ([1]_, [2]_) on grid points.

        Defined as follows:

        .. math::
            \int_G \frac{1}{\alpha - 1} f(x) \bigg(\frac{f(x)}{g(x)}^{q- 1} - 1\bigg) dx,

        where
        :math:`f` is the true probability distribution,
        :math:`g` is the model probability distribution.
        :math:`G` is the grid being integrated on.

        Parameters
        ----------
        density : ndarray(N,)
            The exact density evaluated on the grid points.
        model : ndarray(N,)
            The model density evaluated on the grid points.
        deriv : bool, optional
            Whether to compute the derivative of squared difference w.r.t. model density.

        Returns
        -------
        m : ndarray(N,)
            The Tsallis divergence between density & model on the grid points.
        dm : ndarray(N,)
            The derivative of divergence w.r.t. model density evaluated on the grid points.
            Only returned if `deriv=True`.

        Notes
        -----
        - This does not impose non-negativity of the model density, unlike the Kullback-Leibler.

        - Values of Model density that are less than `mask_value` are masked when used in
          division and then replaced with the value of 0.

        - As :math:`\alpha` parameter tends towards one, then this converges to the
          Kullback-Leibler. This is particularly useful for trust-region methods that don't
          impose strict constraints during the optimization procedure.

        - Care should be taken on :math:`\alpha`. If :math:`\alpha > 1` and it isn't too
          close to 1, then Tsallis divergence upper bounds the Kullback-Leibler and can be useful
          for optimization purposes.

        References
        ----------
        .. [1] Ayers, Paul W. "Information theory, the shape function, and the Hirshfeld atom."
           Theoretical Chemistry Accounts 115.5 (2006): 370-378.

        .. [2] Heidar-Zadeh, Farnaz, Ivan Vinogradov, and Paul W. Ayers. "Hirshfeld partitioning
           from non-extensive entropies." Theoretical Chemistry Accounts 136.4 (2017): 54.

        """
        # check model density
        if not isinstance(model, np.ndarray):
            raise ValueError(
                f"Argument model should be {density.shape} array."
            )
        if not isinstance(density, np.ndarray) or density.ndim != 1:
            raise ValueError("Arguments density should be a 1D numpy array.")
        if model.shape != density.shape:
            raise ValueError(f"Model shape {model.shape} should be the same as density"
                             f" {density.shape}.")
        if not isinstance(deriv, bool):
            raise TypeError(f"Deriv {type(deriv)} should be Boolean type.")

        # compute ratio & replace masked values by 1.0
        ratio = density / np.ma.masked_less_equal(model, self.mask_value)
        ratio = np.ma.filled(ratio, fill_value=0.0)
        value = density * (np.power(ratio, self.alpha - 1.0) - 1.0)
        integrand = value / (self.alpha - 1.0)
        if deriv:
            return integrand, -np.power(ratio, self.alpha)
        return integrand



# -*- coding: utf-8 -*-
# BFit is a Python library for fitting a convex sum of Gaussian
# functions to any probability distribution
#
# Copyright (C) 2020- The QC-Devs Community
#
# This file is part of BFit.
#
# BFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# BFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# ---
r"""Test bfit.density module."""

from bfit.density import SlaterAtoms
from bfit.grid import ClenshawRadialGrid
import numpy as np
from numpy.testing import assert_almost_equal, assert_equal, assert_raises
import scipy


def slater(e, n, r, derivative=False):
    """Calculate single normalized slater function at a given point."""
    norm = np.power(2. * e, n) * np.sqrt(2. * e / scipy.special.factorial(2 * n))
    slater = norm * np.power(r, n - 1) * np.exp(-e * r)

    if derivative:
        return (((n - 1) / r) - e) * slater
    return slater


def test_slater_type_orbital_be():
    r"""Test evaluation of Slater-type orbitals of Beryllium."""
    # load Be atomic wave function
    be = SlaterAtoms("Be")
    # check values of a single orbital at r=1.0
    orbital = be.radial_slater_orbital(np.array([[12.683501]]), np.array([[1]]), np.array([1]))
    assert_almost_equal(orbital, slater(12.683501, 1, 1.0), decimal=6)
    # check values of a single orbital at r=2.0
    orbital = be.radial_slater_orbital(np.array([[0.821620]]), np.array([[2]]), np.array([2]))
    assert_almost_equal(orbital, slater(0.821620, 2, 2.0), decimal=6)
    # check value of tow orbitals at r=1.0 & r=2.0
    exps, nums = np.array([[12.683501], [0.821620]]), np.array([[1], [2]])
    orbitals = be.radial_slater_orbital(exps, nums, np.array([1., 2.]))
    expected = np.array([[slater(exps[0, 0], nums[0, 0], 1.), slater(exps[1, 0], nums[1, 0], 1.)],
                         [slater(exps[0, 0], nums[0, 0], 2.), slater(exps[1, 0], nums[1, 0], 2.)]])
    assert_almost_equal(orbitals, expected, decimal=6)


def test_derivative_slater_type_orbital_be():
    r"""Test derivative of Slater-type orbitals of Beryllium."""
    # load Be atomic wave function
    be = SlaterAtoms("Be")
    # check values of a single orbital at r=1.0
    orbital = be.derivative_radial_slater_type_orbital(np.array([[12.683501]]),
                                                       np.array([[1]]), np.array([1]))
    assert_almost_equal(orbital, slater(12.683501, 1, 1.0, derivative=True), decimal=6)
    # check values of a single orbital at r=2.0
    orbital = be.derivative_radial_slater_type_orbital(np.array([[0.821620]]), np.array([[2]]),
                                                       np.array([2]))
    assert_almost_equal(orbital, slater(0.821620, 2, 2.0, derivative=True), decimal=6)
    # check value of single orbital at r = 0.0
    orbital = be.derivative_radial_slater_type_orbital(np.array([[0.821620]]), np.array([[2]]),
                                                       np.array([0.]))
    assert_almost_equal(orbital, 0.0, decimal=6)
    # check value of tow orbitals at r=1.0 & r=2.0
    exps, nums = np.array([[12.683501], [0.821620]]), np.array([[1], [2]])
    orbitals = be.derivative_radial_slater_type_orbital(exps, nums, np.array([1., 2.]))
    expected = np.array([[slater(exps[0, 0], nums[0, 0], 1., True),
                          slater(exps[1, 0], nums[1, 0], 1., True)],
                         [slater(exps[0, 0], nums[0, 0], 2., True),
                          slater(exps[1, 0], nums[1, 0], 2., True)]])
    assert_almost_equal(orbitals, expected, decimal=6)


def test_positive_definite_kinetic_energy_he():
    r"""Test integral of kinetic energy density of helium against actual value."""
    # load he atomic wave function
    he = SlaterAtoms("he")
    # compute density on an equally distant grid
    grid = ClenshawRadialGrid(4, 30000, 35000)
    energ = he.positive_definite_kinetic_energy(grid.points)
    integral = 4.0 * np.pi * grid.integrate(energ * grid.points**2.0)
    assert np.all(np.abs(integral - he.kinetic_energy) < 1e-5)


def test_positive_definite_kinetic_energy_li():
    r"""Test integral of kinetic energy density of lithium against actual value."""
    # load be atomic wave function
    be = SlaterAtoms("li")
    # compute density on an equally distant grid
    grid = ClenshawRadialGrid(3, 20000, 35000)
    energ = be.positive_definite_kinetic_energy(grid.points)
    integral = 4.0 * np.pi * grid.integrate(energ * grid.points**2.0)
    assert np.all(np.abs(integral - be.kinetic_energy) < 1e-5)


def test_positive_definite_kinetic_energy_c():
    r"""Test integral of kinetic energy density of carbon against actual value."""
    # load be atomic wave function
    be = SlaterAtoms("c")
    # compute density on an equally distant grid
    grid = ClenshawRadialGrid(6, 20000, 35000)
    energ = be.positive_definite_kinetic_energy(grid.points)
    integral = 4.0 * np.pi * grid.integrate(energ * grid.points**2.0)
    assert np.all(np.abs(integral - be.kinetic_energy) < 1e-5)


def test_positive_definite_kinetic_energy_p():
    r"""Test integral of kinetic energy density of phosphorous against actual value."""
    # load be atomic wave function
    be = SlaterAtoms("p")
    # compute density on ClenshawCurtis Grid
    grid = ClenshawRadialGrid(15, 20000, 35000)
    energ = be.positive_definite_kinetic_energy(grid.points)
    integral = 4.0 * np.pi * grid.integrate(energ * grid.points**2.0)
    assert np.all(np.abs(integral - be.kinetic_energy) < 1e-3)


def test_positive_definite_kinetic_energy_ag():
    r"""Test integral of kinetic energy density of silver against actual value."""
    # load c atomic wave function
    adens = SlaterAtoms("ag")
    # compute density on ClenshawCurtis Grid
    grid = ClenshawRadialGrid(47, 20000, 35000)
    energ = adens.positive_definite_kinetic_energy(grid.points)
    integral = 4.0 * np.pi * grid.integrate(energ * grid.points**2.0)
    assert np.all(np.abs(integral - adens.kinetic_energy) < 1e-3)


def test_phi_derivative_lcao_b():
    r"""Test derivative of molecular orbitals of Beryllium."""
    # load Be atomic wave function
    b = SlaterAtoms("b")
    # check the values of the phi_matrix at point 1.0
    phi_matrix = b.phi_matrix(np.array([1]), deriv=True)

    def _slater_deriv(r):
        # compute expected value of 1S
        phi1s = slater(16.109305, 2, r, True) * -0.0005529 + slater(7.628062, 1, r, True) * -0.23501
        phi1s += slater(6.135799, 2, r, True) * -0.1508924 + slater(4.167618, 1, r, True) * -0.64211
        phi1s += slater(2.488602, 1, r, True) * -0.0011507 + slater(1.642523, 2, r, True) * -0.00086
        phi1s += slater(0.991698, 1, r, True) * 0.0004712 + slater(0.787218, 1, r, True) * -0.000232
        # compute expected value of 2S
        phi2s = slater(16.109305, 2, r, True) * -0.0001239 + slater(7.628062, 1, r, True) * 0.012224
        phi2s += slater(6.135799, 2, r, True) * 0.0355967 + slater(4.167618, 1, r, True) * -0.198721
        phi2s += slater(2.488602, 1, r, True) * -0.5378967 + slater(1.642523, 2, r, True) * -0.11997
        phi2s += slater(0.991698, 1, r, True) * 1.4382402 + slater(0.787218, 1, r, True) * 0.0299258
        # compute expected value of 1P
        phi3s = slater(12.135370, 3, r, True) * 0.0000599 + slater(5.508493, 2, r, True) * 0.0113751
        phi3s += slater(3.930298, 3, r, True) * 0.0095096 + slater(2.034395, 2, r, True) * 0.1647518
        phi3s += slater(1.301082, 2, r, True) * 0.3367860 + slater(0.919434, 2, r, True) * 0.4099162
        phi3s += slater(0.787218, 2, r, True) * 0.1329396
        return [phi1s, phi2s, phi3s]
    # check shape & orbital values
    assert_equal(phi_matrix.shape, (1, 3))
    assert_almost_equal(phi_matrix, np.array([_slater_deriv(1)]), decimal=4)
    # check the values of the phi_matrix at point 1.0, 2.0, 3.0
    phi_matrix = b.phi_matrix(np.array([1., 2., 3.]), True)
    # check shape & orbital values
    assert_equal(phi_matrix.shape, (3, 3))
    assert_almost_equal(phi_matrix[0, :], _slater_deriv(1), decimal=4)
    assert_almost_equal(phi_matrix[1, :], _slater_deriv(2.), decimal=4)
    assert_almost_equal(phi_matrix[2, :], _slater_deriv(3.), decimal=4)


def test_coeff_matrix_be():
    r"""Test the coefficients of Beryllium is correctly parsed."""
    # load Be atomic wave function
    be = SlaterAtoms("be")
    # using one _grid point at 1.0
    coeff_1s = np.array([-0.0024917, 0.0314015, 0.0849694, 0.8685562,
                         0.0315855, -0.0035284, -0.0004149, .0012299])[:, None]
    coeff_2s = np.array([0.0004442, -0.0030990, -0.0367056, 0.0138910,
                         -0.3598016, -0.2563459, 0.2434108, 1.1150995])[:, None]
    assert_almost_equal(be.orbitals_coeff["1S"], coeff_1s, decimal=6)
    assert_almost_equal(be.orbitals_coeff["2S"], coeff_2s, decimal=6)


def test_phi_lcao_be():
    r"""Test evaluation of molecular orbitals of beryllium."""
    # load Be atomic wave function
    be = SlaterAtoms("BE")
    # check the values of the phi_matrix at point 1.0
    phi_matrix = be.phi_matrix(np.array([1]))
    # compute expected value of 1S
    phi1s = slater(12.683501, 1, 1) * -0.0024917 + slater(8.105927, 1, 1) * 0.0314015
    phi1s += slater(5.152556, 1, 1) * 0.0849694 + slater(3.472467, 1, 1) * 0.8685562
    phi1s += slater(2.349757, 1, 1) * 0.0315855 + slater(1.406429, 1, 1) * -0.0035284
    phi1s += slater(0.821620, 2, 1) * -0.0004149 + slater(0.786473, 1, 1) * 0.0012299
    # compute expected value of 2S
    phi2s = slater(12.683501, 1, 1) * 0.0004442 + slater(8.105927, 1, 1) * -0.0030990
    phi2s += slater(5.152556, 1, 1) * -0.0367056 + slater(3.472467, 1, 1) * 0.0138910
    phi2s += slater(2.349757, 1, 1) * -0.3598016 - slater(1.406429, 1, 1) * 0.2563459
    phi2s += slater(0.821620, 2, 1) * 0.2434108 + slater(0.786473, 1, 1) * 1.1150995
    # check shape & orbital values
    assert_equal(phi_matrix.shape, (1, 2))
    assert_almost_equal(phi_matrix, np.array([[phi1s, phi2s]]), decimal=6)
    # check the values of the phi_matrix at point 1.0, 2.0, 3.0
    phi_matrix = be.phi_matrix(np.array([1., 2., 3.]))
    # check shape & orbital values
    assert_equal(phi_matrix.shape, (3, 2))
    assert_almost_equal(phi_matrix[0, :], np.array([phi1s, phi2s]), decimal=6)


def test_orbitals_norm_be():
    r"""Test that molecular orbitals are normalized of beryllium."""
    # load Be atomic wave function
    be = SlaterAtoms("be")
    # compute orbital density on an equally distant grid
    grid = np.arange(0.0, 15.0, 0.0001)
    dens = be.phi_matrix(grid)**2
    # check shape
    assert_equal(dens.shape, (grid.size, 2))
    # check orbital normalization
    assert_almost_equal(np.trapz(grid**2 * dens[:, 0], grid), 1.0, decimal=6)
    assert_almost_equal(np.trapz(grid**2 * dens[:, 1], grid), 1.0, decimal=6)


def test_orbitals_norm_ne():
    r"""Test that molecular orbitals are normalized of neon."""
    # load Ne atomic wave function
    ne = SlaterAtoms("ne")
    # compute orbital density on an equally distant grid
    grid = np.arange(0.0, 10.0, 0.0001)
    dens = ne.phi_matrix(grid)**2
    # check shape
    assert_equal(dens.shape, (grid.size, 3))
    # check orbital normalization
    assert_almost_equal(np.trapz(grid**2 * dens[:, 0], grid), 1.0, decimal=6)
    assert_almost_equal(np.trapz(grid**2 * dens[:, 1], grid), 1.0, decimal=6)
    assert_almost_equal(np.trapz(grid**2 * dens[:, 2], grid), 1.0, decimal=6)


def test_orbitals_norm_c():
    r"""Test that molecular orbitals are normalized of carbon."""
    # load C atomic wave function
    c = SlaterAtoms("c")
    # compute orbital density on an equally distant grid
    grid = np.arange(0.0, 15.0, 0.0001)
    dens = c.phi_matrix(grid)**2
    # check shape
    assert_equal(dens.shape, (grid.size, 3))
    # check orbital normalization
    assert_almost_equal(np.trapz(grid**2 * dens[:, 0], grid), 1.0, decimal=6)
    assert_almost_equal(np.trapz(grid**2 * dens[:, 1], grid), 1.0, decimal=6)
    assert_almost_equal(np.trapz(grid**2 * dens[:, 2], grid), 1.0, decimal=6)


def test_atomic_density_be():
    r"""Test integral of atomic densities matches atomic number of beryllium."""
    # load Be atomic wave function
    be = SlaterAtoms("be")
    # compute density on an equally distant grid
    grid = np.arange(0.0, 15.0, 0.0001)
    dens = be.atomic_density(grid, mode="total")
    core = be.atomic_density(grid, mode="core")
    valn = be.atomic_density(grid, mode="valence")
    # check shape
    assert_equal(dens.shape, grid.shape)
    assert_equal(core.shape, grid.shape)
    assert_equal(valn.shape, grid.shape)
    # check dens = core + valence
    assert_almost_equal(dens, core + valn, decimal=6)
    # check number of electrons
    assert_almost_equal(4 * np.pi * np.trapz(grid**2 * dens, grid), 4.0, decimal=6)


def test_atomic_density_ne():
    r"""Test integral of atomic densities matches atomic number of neon."""
    # load Ne atomic wave function
    ne = SlaterAtoms("ne")
    # compute density on an equally distant grid
    grid = np.arange(0.0, 10.0, 0.0001)
    dens = ne.atomic_density(grid, mode="total")
    core = ne.atomic_density(grid, mode="core")
    valn = ne.atomic_density(grid, mode="valence")
    # check shape
    assert_equal(dens.shape, grid.shape)
    assert_equal(core.shape, grid.shape)
    assert_equal(valn.shape, grid.shape)
    # check dens = core + valence
    assert_almost_equal(dens, core + valn, decimal=6)
    # check number of electrons
    assert_almost_equal(4 * np.pi * np.trapz(grid**2 * dens, grid), 10.0, decimal=6)


def test_atomic_density_c():
    r"""Test integral of atomic densities matches atomic number of carbon."""
    # load C atomic wave function
    c = SlaterAtoms("c")
    # compute density on an equally distant grid
    grid = np.arange(0.0, 15.0, 0.0001)
    dens = c.atomic_density(grid, mode="total")
    core = c.atomic_density(grid, mode="core")
    valn = c.atomic_density(grid, mode="valence")
    # check shape
    assert_equal(dens.shape, grid.shape)
    assert_equal(core.shape, grid.shape)
    assert_equal(valn.shape, grid.shape)
    # check dens = core + valence
    assert_almost_equal(dens, core + valn, decimal=6)
    # check number of electrons
    assert_almost_equal(4 * np.pi * np.trapz(grid**2 * dens, grid), 6.0, decimal=6)


def test_atomic_density_h():
    r"""Test integration of atomic density of Hydrogen."""
    h = SlaterAtoms("h")
    grid = np.arange(0.0, 15.0, 0.0001)
    dens = h.atomic_density(grid, mode="total")
    assert_almost_equal((4 * np.pi) * np.trapz(dens * grid**2., grid), 1.0, decimal=6)


def test_atomic_density_h_anion():
    r"""Test integration of atomic density of hydrogen anion."""
    assert_raises(ValueError, SlaterAtoms, "h", cation=True)  # No cations for hydrogen.
    h = SlaterAtoms("h", anion=True)
    grid = np.arange(0.0, 25.0, 0.00001)
    dens = h.atomic_density(grid, mode="total")
    assert_almost_equal((4 * np.pi) * np.trapz(dens * grid ** 2., grid), 2.0, decimal=6)


def test_atomic_density_c_anion_cation():
    r"""Test integration of atomic density of carbon anion and cation."""
    c = SlaterAtoms("c", anion=True)
    grid = np.arange(0.0, 25.0, 0.00001)
    dens = c.atomic_density(grid, mode="total")
    assert_almost_equal((4 * np.pi) * np.trapz(dens * grid ** 2., grid), 7.0, decimal=6)

    c = SlaterAtoms("c", cation=True)
    grid = np.arange(0.0, 25.0, 0.00001)
    dens = c.atomic_density(grid, mode="total")
    assert_almost_equal((4 * np.pi) * np.trapz(dens * grid ** 2., grid), 5.0, decimal=6)


def test_atomic_density_heavy_cs():
    r"""Test integration of atomic density of carbon anion and cation."""
    # These files don't exist.
    assert_raises(ValueError, SlaterAtoms, "cs", cation=True)
    assert_raises(ValueError, SlaterAtoms, "cs", anion=True)

    cs = SlaterAtoms("cs")
    grid = np.arange(0.0, 40.0, 0.0001)
    dens = cs.atomic_density(grid, mode="total")
    assert_almost_equal(4 * np.pi * np.trapz(dens * grid ** 2., grid), 55.0, decimal=5)


def test_atomic_density_heavy_rn():
    r"""Test integration of atomic density of carbon anion and cation."""
    # These files don't exist.
    assert_raises(ValueError, SlaterAtoms, "rn", cation=True)
    assert_raises(ValueError, SlaterAtoms, "rn", anion=True)

    rn = SlaterAtoms("rn")
    grid = np.arange(0.0, 40.0, 0.0001)
    dens = rn.atomic_density(grid, mode="total")
    assert_almost_equal(4 * np.pi * np.trapz(dens * grid ** 2., grid), 86, decimal=5)


def test_kinetic_energy_cation_anion_c():
    r"""Test integral of kinetic energy density of cation/anion of carbon."""
    c = SlaterAtoms("c", cation=True)
    grid = ClenshawRadialGrid(6, 20000, 35000)
    energ = c.positive_definite_kinetic_energy(grid.points)
    integral = 4.0 * np.pi * grid.integrate(energ * grid.points**2.0)
    assert_almost_equal(integral, c.kinetic_energy, decimal=6)

    c = SlaterAtoms("c", anion=True)
    grid = ClenshawRadialGrid(7, 20000, 35000)
    energ = c.positive_definite_kinetic_energy(grid.points)
    integral = 4.0 * np.pi * grid.integrate(energ * grid.points**2.0)
    assert_almost_equal(integral, c.kinetic_energy, decimal=5)


def test_derivative_electron_density_c():
    r"""Test derivative of atomic density of cation of carbon."""
    c = SlaterAtoms("c", cation=True)
    eps = 1e-10
    grid = np.array([0.1, 0.1 + eps, 0.5, 0.5 + eps])
    dens = c.atomic_density(grid)
    actual = c.derivative_density(np.array([0.1, 0.5]))
    desired_0 = (dens[1] - dens[0]) / eps
    desired_1 = (dens[3] - dens[2]) / eps
    assert_almost_equal(actual, np.array([desired_0, desired_1]), decimal=4)


def test_derivative_electron_density_cr():
    r"""Test derivative of atomic density of chromium."""
    cr = SlaterAtoms("cr")
    eps = 2.0e-8
    grid = np.array([0.1 - eps, 0.1, 0.1 + eps, 1. - eps, 1., 1. + eps])
    dens = cr.atomic_density(grid)
    actual = cr.derivative_density(np.array([0.1, 1.]))
    desired_0 = (dens[2] - dens[0]) / (2. * eps)
    desired_1 = (dens[5] - dens[3]) / (2. * eps)
    assert_almost_equal(actual, np.array([desired_0, desired_1]), decimal=4)


def test_kinetic_energy_heavy_element_ce():
    r"""Test integral of kinetic energy of cesium."""
    c = SlaterAtoms("ce")
    grid = ClenshawRadialGrid(55, 10000, 30000)
    energ = c.positive_definite_kinetic_energy(grid.points)
    integral = 4.0 * np.pi * grid.integrate(energ * grid.points**2.0)
    assert_almost_equal(integral, c.kinetic_energy, decimal=2)


def test_raises():
    r"""Test raises error of SlaterAtoms."""
    assert_raises(TypeError, SlaterAtoms, 25)
    assert_raises(TypeError, SlaterAtoms, "be2")
    assert_raises(
        ValueError, SlaterAtoms.radial_slater_orbital,
        np.array([[1]]), np.array([[2]]), np.array([[1.]])
    )
    c = SlaterAtoms("c")
    assert_raises(ValueError, c.atomic_density, np.array([[1.]]), "not total")



# -*- coding: utf-8 -*-
# BFit is a Python library for fitting a convex sum of Gaussian
# functions to any probability distribution
#
# Copyright (C) 2020- The QC-Devs Community
#
# This file is part of BFit.
#
# BFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# BFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# ---
r"""Test bfit.fit module."""

from bfit.fit import KLDivergenceFPI, ScipyFit
from bfit.grid import CubicGrid, UniformRadialGrid
from bfit.measure import KLDivergence, SquaredDifference
from bfit.model import AtomicGaussianDensity, MolecularGaussianDensity
import numpy as np
from numpy.testing import assert_almost_equal, assert_equal, assert_raises


def test_lagrange_multiplier():
    r"""Test lagrange multiplier of KL-SCF."""
    g = UniformRadialGrid(150, 1e-4, 15.0)
    e = np.exp(-g.points)
    kl = KLDivergenceFPI(g, e, None, spherical=False)
    assert_almost_equal(kl.lagrange_multiplier, 1., decimal=8)


def test_goodness_of_fit():
    r"""Test goodness of fit with analytic values."""
    g = UniformRadialGrid(1000, 0.0, 10.0)
    e = np.exp(-g.points)
    m = AtomicGaussianDensity(g.points, num_s=1, num_p=0, normalize=False)
    kl = KLDivergenceFPI(g, e, m, mask_value=0., spherical=True)
    gf = kl.goodness_of_fit(np.array([1.]), np.array([1.]))
    expected = [
        5.56833, 4 * np.pi * 1.60909, 0.128, 4.0 * np.pi * 0.0882922, 4. * np.pi * 17.360
    ]
    assert_almost_equal(expected, gf, decimal=1)


def test_assertion_raises():
    r"""Test assertion raises of all fitting methods."""
    g = UniformRadialGrid(1000, 0.0, 10.0)
    e = np.exp(-g.points)
    m = AtomicGaussianDensity(g.points, num_s=1, num_p=0, normalize=False)
    kl = KLDivergenceFPI(g, e, m, mask_value=0., spherical=True)
    assert_raises(ValueError, kl._update_params, None, None, False, False)
    assert_raises(ValueError, kl.run, np.array([1., 2.]), np.array([1.]))
    assert_raises(ValueError, kl.run, np.array([1.]), np.array([1., 2.]))
    assert_raises(ValueError, kl.run, np.array([1.]), np.array([1.]), False, False)
    assert not kl.run(np.array([1.]), np.array([1.]), maxiter=1)["success"]

    m2 = AtomicGaussianDensity(g.points * 2., num_s=1, num_p=0, normalize=False)
    measure = KLDivergence()
    assert_raises(ValueError, ScipyFit, g, e, m2)
    assert_raises(ValueError, ScipyFit, g, np.append(e, 2.), m)
    assert_raises(ValueError, ScipyFit, g, e, m, measure, "non slsqp")
    assert_raises(TypeError, ScipyFit, g, e, m, "not kl")
    gb = ScipyFit(g, e, m, measure=measure)
    assert_raises(ValueError, gb.run, [], [], False, False)
    assert_raises(ValueError, gb.evaluate_model, [], ("not fixed", 2))

    # Test giving a grid class with no points returns error
    class GridNoPoints:
        def integrate(self):
            pass
    assert_raises(AttributeError, ScipyFit, GridNoPoints(), e, KLDivergence())

    # Test giving grid class with no integrate method returns error.
    class GridNoIntegrate:
        def __init__(self):
            self.points = g.points
    assert_raises(AttributeError, ScipyFit, GridNoIntegrate(), e, KLDivergence())


def test_run_normalized_s_gaussian():
    r"""Test KL-SCF against normalized s-type Gaussian."""
    # density is normalized 1s orbital with exponent=1.0
    g = UniformRadialGrid(150, 0.0, 15.0)
    e = (1. / np.pi)**1.5 * np.exp(-g.points**2.)
    model = AtomicGaussianDensity(g.points, num_s=1, num_p=0, normalize=True)
    kl = KLDivergenceFPI(g, e, model, spherical=True)

    # fit density with initial coeff=1. & expon=1.
    res = kl.run(np.array([1.]), np.array([1.]), True, True, 500, 1.e-4, 1.e-4, 1.e-4)
    # check optimized coeffs & expons
    assert_almost_equal(np.array([1.]), res["coeffs"], decimal=8)
    assert_almost_equal(np.array([1.]), res["exps"], decimal=8)
    # check value of optimized objective function & fitness measure
    assert_almost_equal(0., res["fun"][-1], decimal=8)
    assert_almost_equal(1., res["performance"][-1, 0], decimal=8)
    assert_almost_equal(0., res["performance"][-1, 1:], decimal=8)

    # fit density with initial coeff=0.5 & expon=0.5
    res = kl.run(np.array([0.5]), np.array([0.5]), True, True, 500, 1.e-4, 1.e-4, 1.e-4)
    # check optimized coeffs & expons
    assert_almost_equal(np.array([1.]), res["coeffs"], decimal=8)
    assert_almost_equal(np.array([1.]), res["exps"], decimal=8)
    # check value of optimized objective function & fitness measure
    assert_almost_equal(0., res["fun"][-1], decimal=8)
    assert_almost_equal(1., res["performance"][-1, 0], decimal=8)
    assert_almost_equal(0., res["performance"][-1, 1:], decimal=8)

    # fit density with initial coeff=0.1 & expon=10.
    res = kl.run(np.array([0.1]), np.array([10.]), True, True, 500, 1.e-4, 1.e-4, 1.e-4)
    # check optimized coeffs & expons
    assert_almost_equal(np.array([1.]), res["coeffs"], decimal=8)
    assert_almost_equal(np.array([1.]), res["exps"], decimal=8)
    # check value of optimized objective function
    assert_almost_equal(0., res["fun"][-1], decimal=8)

    # fit density with initial coeff=20. & expon=0.01.
    res = kl.run(np.array([20.]), np.array([0.01]), True, True, 500, 1.e-4, 1.e-4, 1.e-4)
    # check optimized coeffs & expons
    assert_almost_equal(np.array([1.]), res["coeffs"], decimal=8)
    assert_almost_equal(np.array([1.]), res["exps"], decimal=8)
    # check value of optimized objective function
    assert_almost_equal(0., res["fun"][-1], decimal=8)


def test_kl_scf_update_coeffs_2s_gaussian():
    r"""Test KL-SCF method for updating coefficients of two s-type Gaussians."""
    # actual density is a 1s Slater function
    grid = UniformRadialGrid(150, 0.0, 15.0)
    c, e = np.array([5., 2.]), np.array([10., 3.])
    dens = np.exp(-grid.points, dtype=np.float64)
    # model density is a normalized 2s Gaussian basis
    model = AtomicGaussianDensity(grid.points, num_s=2, num_p=0, normalize=True)
    # test updating coeffs
    kl = KLDivergenceFPI(grid, dens, model, spherical=False)
    new_coeffs, new_expons = kl._update_params(c, e, True, False)
    # compute model density
    approx = c[0] * (e[0] / np.pi)**1.5 * np.exp(-e[0] * grid.points**2)
    approx += c[1] * (e[1] / np.pi)**1.5 * np.exp(-e[1] * grid.points**2)
    # compute expected coeffs
    coeffs = c * (e / np.pi)**1.5
    ratio = dens / approx
    coeffs[0] *= grid.integrate(ratio * np.exp(-e[0] * grid.points**2))
    coeffs[1] *= grid.integrate(ratio * np.exp(-e[1] * grid.points**2))
    assert_almost_equal(new_expons, e, decimal=6)
    assert_almost_equal(new_coeffs, coeffs, decimal=6)


def test_kl_scf_update_params_2s_gaussian():
    r"""Test KL-SCF method for updating exponents of two s-type Gaussians."""
    # actual density is a 1s Slater function
    grid = UniformRadialGrid(1000, 0.0, 20.0)
    points = grid.points
    spherical = 4.0 * np.pi * points**2.0
    c, e = np.array([5., 2.], dtype=np.float64), np.array([10., 3.], dtype=np.float64)
    dens = np.exp(-points, dtype=np.float64)

    # model density is a normalized 2s Gaussian basis
    model = AtomicGaussianDensity(points, num_s=2, num_p=0, normalize=True)
    # test updating coeffs
    kl = KLDivergenceFPI(grid, dens, model, spherical=True)
    new_coeffs, new_expons = kl._update_params(c, e, False, True)
    # compute model density
    approx = c[0] * (e[0] / np.pi)**1.5 * np.exp(-e[0] * points**2)
    approx += c[1] * (e[1] / np.pi)**1.5 * np.exp(-e[1] * points**2)
    # compute expected expons
    expons = 1.5 * np.ones(2)
    ratio = np.ma.filled(dens / np.ma.array(approx), 0.)
    expons[0] *= grid.integrate(np.exp(-e[0] * points**2) * ratio * spherical)
    expons[1] *= grid.integrate(np.exp(-e[1] * points**2) * ratio * spherical)
    expons[0] /= grid.integrate((np.exp(-e[0] * points**2) * points**2) * ratio * spherical)
    expons[1] /= grid.integrate((np.exp(-e[1] * points**2) * points**2) * ratio * spherical)
    assert_almost_equal(new_coeffs, c, decimal=6)
    assert_almost_equal(new_expons, expons, decimal=6)


def test_kl_scf_update_params_1s1p_gaussian():
    r"""Test KL-SCF method against one s-type and one p-type Gaussians."""
    # actual density is a 1s Slater function
    grid = UniformRadialGrid(150, 0.0, 15.0)
    points = grid.points
    spherical = 4.0 * np.pi * points**2.0
    c, e = np.array([1., 2.]), np.array([3., 4.])
    dens = np.exp(-grid.points)
    # model density is a normalized 2s Gaussian basis
    model = AtomicGaussianDensity(points, num_s=1, num_p=1, normalize=True)
    # compute model density
    approx = c[0] * (e[0] / np.pi)**1.5 * np.exp(-e[0] * points**2)
    approx += c[1] * 2. * e[1]**2.5 * points**2 * np.exp(-e[1] * points**2) / (3 * np.pi**1.5)
    # check model.evaluate
    assert_almost_equal(approx, model.evaluate(c, e), decimal=6)
    # test updating coeffs
    kl = KLDivergenceFPI(grid, dens, model, mask_value=0., spherical=True)
    new_coeffs, new_expons = kl._update_params(c, e, update_coeffs=True, update_expons=False)
    coeffs = c * np.array([(e[0] / np.pi)**1.5, 2 * e[1]**2.5 / (3 * np.pi**1.5)])
    coeffs[0] *= grid.integrate(dens * np.exp(-e[0] * points**2) / approx * spherical)
    coeffs[1] *= grid.integrate(dens * np.exp(-e[1] * points**2) * points**2 / approx * spherical)
    assert_almost_equal(new_expons, e, decimal=6)
    assert_almost_equal(new_coeffs, coeffs, decimal=6)
    # test updating expons
    new_coeffs, new_expons = kl._update_params(c, e, update_coeffs=False, update_expons=True)
    expons = np.array([1.5, 2.5])
    expons[0] *= grid.integrate(dens * np.exp(-e[0] * points**2) / approx * spherical)
    expons[1] *= grid.integrate(dens * points**2 * np.exp(-e[1] * points**2) / approx * spherical)
    expons[0] /= grid.integrate(dens * points**2 * np.exp(-e[0] * points**2) / approx * spherical)
    expons[1] /= grid.integrate(dens * points**4 * np.exp(-e[1] * points**2) / approx * spherical)
    assert_almost_equal(new_coeffs, c, decimal=6)
    assert_almost_equal(new_expons, expons, decimal=6)
    # test updating coeffs & expons
    new_coeffs, new_expons = kl._update_params(c, e, update_coeffs=True, update_expons=True)
    assert_almost_equal(new_coeffs, coeffs, decimal=6)
    assert_almost_equal(new_expons, expons, decimal=6)


def test_kl_scf_update_params_3d_molecular_dens_1s_1s_gaussian():
    r"""Test KL-SCF method in three dimensions for one s-type and one p-type Gaussian."""
    # actual density is a 1s Gaussian at origin
    axes = np.array([[0.1, 0.0, 0.0], [0.0, 0.1, 0.0], [0.0, 0.0, 0.1]])
    grid = CubicGrid(np.array([-2.0, -2.0, -2.0]), axes, (40, 40, 40))
    dens = np.exp(-np.sum(grid.points ** 2., axis=1))
    # model density is a normalized 1s Gassuain on each center
    coord = np.array([[0., 0., 0.], [0., 0., 1.]])
    c, e = np.array([1., 2.]), np.array([3., 4.])
    model = MolecularGaussianDensity(grid.points, coord, np.array([[1, 0], [1, 0]]), True)
    kl = KLDivergenceFPI(grid, dens, model)
    # compute expected updated coeffs
    dist1 = np.sum((grid.points - coord[0])**2, axis=1)
    dist2 = np.sum((grid.points - coord[1])**2, axis=1)
    approx = c[0] * (e[0] / np.pi)**1.5 * np.exp(-e[0] * dist1)
    approx += c[1] * (e[1] / np.pi)**1.5 * np.exp(-e[1] * dist2)
    expected_coeffs = c * (e / np.pi)**1.5
    expected_coeffs[0] *= grid.integrate(dens * np.exp(-e[0] * dist1) / approx)
    expected_coeffs[1] *= grid.integrate(dens * np.exp(-e[1] * dist2) / approx)
    # check updated coeffs
    coeffs, expons = kl._update_params(c, e, update_coeffs=True, update_expons=False)
    assert_almost_equal(expons, e, decimal=6)
    assert_almost_equal(coeffs, expected_coeffs, decimal=6)
    # compute expected updated expons
    expected_expons = 1.5 * (e / np.pi)**1.5
    expected_expons[0] *= grid.integrate(dens * np.exp(-e[0] * dist1) / approx)
    expected_expons[1] *= grid.integrate(dens * np.exp(-e[1] * dist2) / approx)
    denoms = (e / np.pi)**1.5
    denoms[0] *= grid.integrate(dens * dist1 * np.exp(-e[0] * dist1) / approx)
    denoms[1] *= grid.integrate(dens * dist2 * np.exp(-e[1] * dist2) / approx)
    expected_expons /= denoms
    # check updated expons
    coeffs, expons = kl._update_params(c, e, update_coeffs=False, update_expons=True)
    assert_almost_equal(coeffs, c, decimal=6)
    assert_almost_equal(expons, expected_expons, decimal=6)


def test_kl_scf_run_3d_molecular_dens_1s_1p_gaussian():
    r"""Test KL-SCF on 3D Gaussian example with 1 s-type and 1 p-type Gaussians."""
    # make cubic grid
    axes = np.array([[0.3, 0.0, 0.0], [0.0, 0.3, 0.0], [0.0, 0.0, 0.3]])
    grid = CubicGrid(np.array([-10.0, -10.0, -10.0]), axes, (65, 65, 65))
    # actual density is a 1s Gaussian at x=0.5 & 1p Gaussian at y=0.25
    c, e = np.array([0.53, 2.07]), np.array([0.67, 1.92])
    coord = np.array([[0.5, 0., 0.], [0., 0.25, 0.]])
    dist1 = np.sum((grid.points - coord[0])**2., axis=1)
    dist2 = np.sum((grid.points - coord[1])**2., axis=1)
    dens1 = c[0] * (e[0] / np.pi)**1.5 * np.exp(-e[0] * dist1)
    dens2 = c[1] * 2. * e[1]**2.5 * dist2 * np.exp(-e[1] * dist2) / (3. * np.pi**1.5)
    # model density is a normalized 1s Gaussian on each center
    model = MolecularGaussianDensity(grid.points, coord, np.array([[1, 0], [0, 1]]), True)
    # check model norm
    assert_almost_equal(grid.integrate(model.evaluate(c, e, False)), np.sum(c), decimal=8)
    # initial coeffs & expons for optimization
    cs0, es0 = np.array([0.25, 0.01]), np.array([0.5, 2.2])
    # optimize coeffs
    kl = KLDivergenceFPI(grid, dens1 + dens2, model, mask_value=0.)
    result = kl.run(cs0, e, True, False, 100, 1.e-8, 1.e-8, 1e-15)
    assert_equal(result["success"], True)
    assert_almost_equal(result["coeffs"], c, decimal=6)
    assert_almost_equal(result["exps"], e, decimal=6)
    assert_almost_equal(result["fun"][-1], 0., decimal=6)
    # optimize expons
    result = kl.run(c, es0, False, True, 100, 1.e-8, 1.e-8, 1.e-15)
    assert_equal(result["success"], True)
    assert_almost_equal(result["coeffs"], c, decimal=6)
    assert_almost_equal(result["exps"], e, decimal=6)
    assert_almost_equal(result["fun"][-1], 0., decimal=6)
    # optimize coeffs & expons
    result = kl.run(cs0, es0, True, True, 150, 1.e-7, 1.e-7, 1.e-15)
    assert_equal(result["success"], True)
    assert_almost_equal(result["coeffs"], c, decimal=6)
    assert_almost_equal(result["exps"], e, decimal=6)
    assert_almost_equal(result["fun"][-1], 0., decimal=6)


def test_kl_fit_unnormalized_dens_normalized_1s_gaussian():
    r"""Test ScipyFit against normalized gaussian model."""
    # density is normalized 1s orbital with exponent=1.0
    grid = UniformRadialGrid(150, 0.0, 15.0)
    # density is normalized 1s gaussian
    dens = 1.57 * np.exp(-0.51 * grid.points**2.)
    model = AtomicGaussianDensity(grid.points, num_s=1, num_p=0, normalize=True)
    measure = KLDivergence()
    kl = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    # expected coeffs & expons
    expected_cs = np.array([1.57 / (0.51 / np.pi)**1.5])
    expected_es = np.array([0.51])
    # initial coeff=1.57 & expon=0.51
    result = kl.run(np.array([1.57]), np.array([0.51]), True, True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=8)
    assert_almost_equal(expected_es, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., 0.]), result["jacobian"], decimal=6)
    # initial coeff=0.1 & expon=0.1
    result = kl.run(np.array([0.1]), np.array([0.1]), True, True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=8)
    assert_almost_equal(expected_es, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., 0.]), result["jacobian"], decimal=6)
    # initial coeff=5.0 & expon=15.
    result = kl.run(np.array([5.0]), np.array([15.]), True, True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=8)
    assert_almost_equal(expected_es, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., 0.]), result["jacobian"], decimal=6)
    # initial coeff=0.8 & expon=0.51, opt coeffs
    result = kl.run(np.array([0.5]), np.array([0.51]), True, False)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=8)
    assert_almost_equal(expected_es, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1.]), result["jacobian"], decimal=6)


def test_kl_fit_normalized_dens_unnormalized_1s_gaussian():
    r"""Test ScipyFit against un-normalized Gaussian model."""
    # density is normalized 1s gaussian
    grid = UniformRadialGrid(200, 0.0, 15.0)
    dens = 2.06 * (0.88 / np.pi)**1.5 * np.exp(-0.88 * grid.points**2.)
    # un-normalized 1s basis function
    model = AtomicGaussianDensity(grid.points, num_s=1, num_p=0, normalize=False)
    measure = KLDivergence()
    kl = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    # expected coeffs & expons
    expected_cs = np.array([2.06]) * (0.88 / np.pi)**1.5
    expected_es = np.array([0.88])
    # initial coeff=2.6 & expon=0.001
    result = kl.run(np.array([2.6]), np.array([0.001]), opt_coeffs=True, opt_expons=True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=7)
    assert_almost_equal(expected_es, result["exps"], decimal=7)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=1. & expon=0.88, opt coeffs
    result = kl.run(np.array([1.0]), np.array([0.88]), opt_coeffs=True, opt_expons=False)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=8)
    assert_almost_equal(expected_es, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1.]) / (0.88 / np.pi)**1.5, result["jacobian"], decimal=6)
    # initial coeff=10. & expon=0.88, opt coeffs
    result = kl.run(np.array([10.]), np.array([0.88]), opt_coeffs=True, opt_expons=False)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=8)
    assert_almost_equal(expected_es, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1.]) / (0.88 / np.pi)**1.5, result["jacobian"], decimal=6)
    # initial coeff=expected_cs & expon=expected_es, opt expons
    result = kl.run(expected_cs, expected_es, opt_coeffs=False, opt_expons=True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=8)
    assert_almost_equal(expected_es, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=expected_cs & expon=5.0, opt expons
    result = kl.run(expected_cs, np.array([5.0]), opt_coeffs=False, opt_expons=True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=8)
    assert_almost_equal(expected_es, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)


def test_kl_fit_normalized_dens_normalized_1s_gaussian():
    r"""Test ScipyFiy againstnormalized Gaussian model."""
    # density is normalized 1s gaussian
    grid = UniformRadialGrid(150, 0.0, 15.0)
    dens = 2.06 * (0.88 / np.pi)**1.5 * np.exp(-0.88 * grid.points**2.)
    # normalized 1s basis function
    model = AtomicGaussianDensity(grid.points, num_s=1, num_p=0, normalize=True)
    measure = KLDivergence()
    kl = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    # expected coeffs & expons
    expected_cs = np.array([2.06])
    expected_es = np.array([0.88])
    # initial coeff=2.6 & expon=0.001
    result = kl.run(np.array([2.6]), np.array([0.001]), opt_coeffs=True, opt_expons=True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(expected_es, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., 0.]), result["jacobian"], decimal=6)
    # initial coeff=1. & expon=0.88, opt coeffs
    result = kl.run(np.array([1.0]), np.array([0.88]), opt_coeffs=True, opt_expons=False)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(expected_es, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1.]), result["jacobian"], decimal=6)
    # initial coeff=10. & expon=0.88, opt coeffs
    result = kl.run(np.array([10.]), np.array([0.88]), opt_coeffs=True, opt_expons=False)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(expected_es, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1.]), result["jacobian"], decimal=6)
    # initial coeff=expected_cs & expon=expected_es, opt expons
    result = kl.run(expected_cs, expected_es, opt_coeffs=False, opt_expons=True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(expected_es, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([0.]), result["jacobian"], decimal=6)


def test_kl_fit_normalized_dens_unnormalized_2p_gaussian():
    r"""Test ScipyFiy against un-normalized p-type Gaussian."""
    # density is normalized 2p orbitals
    grid = UniformRadialGrid(150, 1e-15, 15.0)
    points = grid.points
    cs0 = np.array([0.76, 3.09])
    es0 = np.array([2.01, 0.83])
    dens = cs0[0] * es0[0]**2.5 * points**2 * np.exp(-es0[0] * points**2)
    dens += cs0[1] * es0[1]**2.5 * points**2 * np.exp(-es0[1] * points**2)
    dens *= 2. / (3. * np.pi**1.5)
    # un-normalized 2p functions
    model = AtomicGaussianDensity(points, num_s=0, num_p=2, normalize=False)
    measure = KLDivergence()
    kl = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    # expected coeffs & expons
    expected_cs = cs0 * es0**2.5 * 2. / (3. * np.pi**1.5)
    expected_es = es0
    # initial coeff=[1.5, 0.1] & expon=[4., 0.001]
    result = kl.run(np.array([1.5, 0.1]), np.array([4.0, 0.001]), True, True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(expected_es, result["exps"], decimal=5)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=cs0 & expon=es0, opt coeffs
    result = kl.run(cs0, es0, opt_coeffs=True, opt_expons=False)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(expected_es, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=[1.0, 1.0] & expon=es0, opt coeffs
    result = kl.run(np.array([1.0, 1.0]), es0, opt_coeffs=True, opt_expons=False)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(expected_es, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=expected_cs & expon=[4.5, 1.0], opt expons
    result = kl.run(expected_cs, np.array([4.5, 1.0]), opt_coeffs=False, opt_expons=True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(expected_es, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)


def test_kl_fit_normalized_dens_normalized_2p_gaussian():
    r"""Test ScipyFiy against normalized p-type Gaussian model."""
    # density is normalized 2p orbitals
    grid = UniformRadialGrid(150, 1e-15, 15.0)

    points = grid.points
    cs0 = np.array([0.76, 3.09])
    es0 = np.array([2.01, 0.83])
    dens = cs0[0] * es0[0]**2.5 * points**2 * np.exp(-es0[0] * points**2)
    dens += cs0[1] * es0[1]**2.5 * points**2 * np.exp(-es0[1] * points**2)
    dens *= 2. / (3. * np.pi ** 1.5)
    # normalized 2p functions
    model = AtomicGaussianDensity(points, num_s=0, num_p=2, normalize=True)
    measure = KLDivergence()
    kl = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    # initial coeff=cs0 & expon=es0, opt coeffs
    result = kl.run(cs0, es0, True, False)
    assert_almost_equal(cs0, result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., -1.]), result["jacobian"], decimal=6)
    # initial coeff=[2.5, 0.5] & expon=[2.0, 1.9]
    result = kl.run(np.array([2.5, 0.5]), np.array([2.0, 1.9]), True, True)
    assert_almost_equal(np.sort(cs0), np.sort(result["coeffs"]), decimal=6)
    assert_almost_equal(np.sort(es0), np.sort(result["exps"]), decimal=5)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., -1., 0., 0.]), result["jacobian"], decimal=6)
    # initial coeff=[1.0, 1.0] & expon=es0, opt coeffs
    result = kl.run(np.array([1.0, 1.0]), es0, True, False)
    assert_almost_equal(cs0, result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., -1.]), result["jacobian"], decimal=6)
    # initial coeff=cs0 & expon=es0, opt coeffs
    result = kl.run(cs0, es0, True, False)
    assert_almost_equal(cs0, result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., -1.]), result["jacobian"], decimal=6)


def test_kl_fit_normalized_dens_normalized_1s2p_gaussian():
    r"""Test ScipyFit with KL against normalized one s-type and 2 p-type Gaussians."""
    # density is normalized 1s + 2p gaussians
    grid = UniformRadialGrid(150, 0.0, 15.0)
    points = grid.points
    cs0 = np.array([1.52, 0.76, 3.09])
    es0 = np.array([0.50, 2.01, 0.83])
    dens = cs0[0] * (es0[0] / np.pi)**1.5 * np.exp(-es0[0] * points**2.)
    dens += cs0[1] * 2 * es0[1]**2.5 * points**2 * np.exp(-es0[1] * points**2) / (3. * np.pi**1.5)
    dens += cs0[2] * 2 * es0[2]**2.5 * points**2 * np.exp(-es0[2] * points**2) / (3. * np.pi**1.5)
    # un-normalized 1s + 2p functions
    model = AtomicGaussianDensity(points, num_s=1, num_p=2, normalize=True)
    measure = KLDivergence()
    kl = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    # initial coeff=1. & expon=1.
    result = kl.run(np.array([1., 1., 1.]), np.array([1., 1., 1.]), True, True)
    assert_almost_equal(np.sort(cs0), np.sort(result["coeffs"]), decimal=5)
    assert_almost_equal(np.sort(es0), np.sort(result["exps"]), decimal=5)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., -1., -1., 0., 0., 0.]), result["jacobian"], decimal=6)
    # initial coeff=[0.1, 0.6, 7.] & expon=[1., 0.9, 1.0]
    result = kl.run(np.array([0.1, 0.6, 7.]), np.array([1., 0.9, 1.0]), True, True)
    assert_almost_equal(np.sort(cs0), np.sort(result["coeffs"]), decimal=5)
    assert_almost_equal(np.sort(es0), np.sort(result["exps"]), decimal=5)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., -1., -1., 0., 0., 0.]), result["jacobian"], decimal=6)
    # initial coeff=[1., 5., 0.] & expon=es0, opt coeffs
    result = kl.run(np.array([1., 5., 0.]), es0, True, False)
    assert_almost_equal(np.sort(cs0), np.sort(result["coeffs"]), decimal=5)
    assert_almost_equal(np.sort(es0), np.sort(result["exps"]), decimal=5)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([-1., -1., -1.]), result["jacobian"], decimal=6)
    # initial coeff=cs0 & expon=es0, opt expons
    result = kl.run(cs0, es0, False, True)
    assert_almost_equal(np.sort(cs0), np.sort(result["coeffs"]), decimal=5)
    assert_almost_equal(np.sort(es0), np.sort(result["exps"]), decimal=5)
    assert_almost_equal(0., result["fun"], decimal=8)
    assert_almost_equal(np.array([0., 0., 0.]), result["jacobian"], decimal=6)


def test_kl_fit_unnormalized_1d_molecular_dens_unnormalized_1s_1s_gaussian():
    r"""Test ScipyFit of 3D un-normalized Gaussian model."""
    # density is normalized 1s + 1s gaussians
    grid = UniformRadialGrid(150, 0.0, 15.0)
    points = grid.points
    cs0 = np.array([1.52, 2.67])
    es0 = np.array([0.31, 0.41])
    coords = np.array([[0.], [1.]])
    # compute density on each center
    dens1 = cs0[0] * np.exp(-es0[0] * (points - coords[0])**2.)
    dens2 = cs0[1] * np.exp(-es0[1] * (points - coords[1])**2.)
    # un-normalized 1s + 1s functions
    model = MolecularGaussianDensity(points, coords, np.array([[1, 0], [1, 0]]), False)
    # fit total density
    measure = KLDivergence()
    kl = ScipyFit(grid, dens1 + dens2, model, measure=measure, method="slsqp", spherical=True)
    # initial coeff=1. & expon=1.
    result = kl.run(np.ones(2), np.ones(2), True, True)
    assert_almost_equal(cs0, result["coeffs"], decimal=4)
    assert_almost_equal(es0, result["exps"], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=1. & expon=1.
    result = kl.run(np.array([5.45, 0.001]), es0, True, False, disp=True)
    assert_almost_equal(cs0, result["coeffs"], decimal=5)
    assert_almost_equal(es0, result["exps"], decimal=5)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=1. & expon=1.
    result = kl.run(cs0, np.array([5.45, 0.001]), False, True)
    assert_almost_equal(cs0, result["coeffs"], decimal=5)
    assert_almost_equal(es0, result["exps"], decimal=5)
    assert_almost_equal(0., result["fun"], decimal=8)
    # fit 1s density on center 1
    measure = KLDivergence()
    kl = ScipyFit(grid, dens1, model, measure=measure, method="slsqp", spherical=True)
    # initial coeff=1. & expon=1.
    result = kl.run(np.ones(2), np.ones(2), True, True)
    assert_almost_equal(np.array([result["coeffs"][0], 0.]), result["coeffs"], decimal=4)
    assert_almost_equal(es0[0], result["exps"][0], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=1. & expon=1.
    result = kl.run(np.array([6.79, 8.51]), es0, True, False)
    assert_almost_equal(np.array([result["coeffs"][0], 0.]), result["coeffs"], decimal=4)
    assert_almost_equal(es0, result["exps"], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)
    # initial coeff=1. & expon=1.
    result = kl.run(np.array([1.52, 0.0]), np.array([3.0, 4.98]), False, True)
    assert_almost_equal(np.array([result["coeffs"][0], 0.]), result["coeffs"], decimal=4)
    assert_almost_equal(es0[0], result["exps"][0], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)


def test_kl_fit_unnormalized_1d_molecular_dens_unnormalized_1s_1p_gaussian():
    r"""Test ScipyFit against un-normalized 1S and 1P-type Gaussians."""
    # density is normalized 1s + 1s gaussians
    grid = UniformRadialGrid(150, 0.0, 15.0)
    points = grid.points
    cs0 = np.array([1.52, 2.67])
    es0 = np.array([0.31, 0.41])
    coords = np.array([[0.], [1.]])
    # compute density of each center
    dens_s = cs0[0] * (es0[0] / np.pi)**1.5 * np.exp(-es0[0] * (points - coords[0])**2.)
    dens_p = cs0[1] * (points - coords[1])**2 * np.exp(-es0[1] * (points - coords[1])**2.)
    dens_p *= (2. * es0[1]**2.5 / (3. * np.pi**1.5))
    # un-normalized 1s + 1p functions
    model = MolecularGaussianDensity(points, coords, np.array([[1, 0], [0, 1]]), True)
    # fit total density
    measure = KLDivergence()
    kl = ScipyFit(grid, dens_s + dens_p, model, measure=measure, method="slsqp", spherical=True)
    # opt. coeffs & expons
    result = kl.run(np.ones(2), np.array([0.1, 0.2]), True, True, tol=1e-20, disp=True)
    assert_almost_equal(cs0, result["coeffs"], decimal=4)
    assert_almost_equal(es0, result["exps"], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)
    # opt. coeffs, initial expon=es0
    result = kl.run(np.array([5.91, 7.01]), es0, True, False)
    assert_almost_equal(cs0, result["coeffs"], decimal=4)
    assert_almost_equal(es0, result["exps"], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)
    # opt. expons, initial coeff=cs0
    result = kl.run(cs0, np.array([5.91, 7.01]), False, True)
    assert_almost_equal(cs0, result["coeffs"], decimal=4)
    assert_almost_equal(es0, result["exps"], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)
    # fit 1s density on center 1
    measure = KLDivergence()
    kl = ScipyFit(grid, dens_s, model, measure=measure, method="slsqp", spherical=True)
    # opt. coeffs & expons
    result = kl.run(np.ones(2), np.ones(2), True, True)
    assert_almost_equal(np.array([cs0[0], 0.]), result["coeffs"], decimal=4)
    assert_almost_equal(es0[0], result["exps"][0], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)
    # # fit 1p density on center 2
    measure = KLDivergence(mask_value=1e-12)
    kl = ScipyFit(grid, dens_p, model, measure=measure, method="slsqp", spherical=True)
    # opt. expons
    result = kl.run(np.array([0., cs0[1]]), np.ones(2), False, True)
    assert_almost_equal(np.array([0., cs0[1]]), result["coeffs"], decimal=4)
    assert_almost_equal(es0[1], result["exps"][1], decimal=4)
    assert_almost_equal(0., result["fun"], decimal=8)


def test_ls_fit_normalized_dens_normalized_1s_gaussian():
    r"""Test ScipyFit of least-squares against normalized Gaussian model."""
    # density is normalized 1s orbital with exponent=1.0
    grid = UniformRadialGrid(200, 0.0, 15.0)
    # actual density is a normalized 1s gaussian
    cs0, es0 = np.array([1.57]), np.array([0.51])
    dens = 1.57 * (0.51 / np.pi)**1.5 * np.exp(-0.51 * grid.points**2.)
    # model density is a normalized 1s Gaussian
    model = AtomicGaussianDensity(grid.points, num_s=1, num_p=0, normalize=True)
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    # opt. coeffs & expons
    result = ls.run(np.array([0.1]), np.array([3.5]), True, True)
    assert_almost_equal(cs0, result["coeffs"], decimal=8)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # opt. coeffs
    result = ls.run(np.array([10.]), np.array([0.51]), True, False)
    assert_almost_equal(cs0, result["coeffs"], decimal=8)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # model density is two normalized 1s Gaussian
    model = AtomicGaussianDensity(grid.points, num_s=2, num_p=0, normalize=True)
    ls = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    # opt. coeffs & expons
    result = ls.run(np.array([0.1, 3.0]), np.array([5.1, 7.8]), True, True)
    assert_almost_equal(np.array([1.57, 0.]), result["coeffs"], decimal=6)
    assert_almost_equal(np.array([0.51, 0.]), result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)


def test_ls_fit_normalized_dens_normalized_2s_gaussian():
    r"""Test ScipyFit least-squares against two normalized s-type Gaussians."""
    # density is normalized 1s orbital with exponent=1.0
    grid = UniformRadialGrid(300, 0.0, 15.0)
    # actual density is a normalized 1s gaussian
    cs0 = np.array([1.57, 0.12])
    es0 = np.array([0.45, 1.29])
    # evaluate normalized 8s density
    dens = cs0[0] * (es0[0] / np.pi)**1.5 * np.exp(-es0[0] * grid.points**2.)
    dens += cs0[1] * (es0[1] / np.pi)**1.5 * np.exp(-es0[1] * grid.points**2.)
    # check norm of density
    assert_almost_equal(
        grid.integrate(dens * 4.0 * np.pi * grid.points**2.0), np.sum(cs0), decimal=6
    )
    # model density is a normalized 2s Gaussian
    model = AtomicGaussianDensity(grid.points, num_s=2, num_p=0, normalize=True)
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    initial_cs = np.array([0.57, 0.98])
    initial_es = np.array([1.67, 0.39])
    # opt. coeffs & expons
    result = ls.run(initial_cs, initial_es, True, True)
    assert_almost_equal(np.sort(cs0), np.sort(result["coeffs"]), decimal=6)
    assert_almost_equal(np.sort(es0), np.sort(result["exps"]), decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # opt. coeffs
    result = ls.run(initial_cs, es0, True, False)
    assert_almost_equal(cs0, result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=8)
    assert_almost_equal(0., result["fun"], decimal=8)


def test_ls_fit_normalized_dens_normalized_5s_gaussian():
    r"""Test ScipyFiy least-squares against 5 s-type, normalized Gaussians."""
    # density is normalized 1s orbital with exponent=1.0
    grid = UniformRadialGrid(300, 0.0, 15.0)
    # actual density is a normalized 5s gaussian
    cs0 = np.array([1.57, 0.12, 3.67, 0.97, 5.05])
    es0 = np.array([0.45, 1.29, 1.25, 20.1, 10.5])
    # evaluate normalized 5s density
    dens = cs0[0] * (es0[0] / np.pi)**1.5 * np.exp(-es0[0] * grid.points**2)
    dens += cs0[1] * (es0[1] / np.pi)**1.5 * np.exp(-es0[1] * grid.points**2)
    dens += cs0[2] * (es0[2] / np.pi)**1.5 * np.exp(-es0[2] * grid.points**2)
    dens += cs0[3] * (es0[3] / np.pi)**1.5 * np.exp(-es0[3] * grid.points**2)
    dens += cs0[4] * (es0[4] / np.pi)**1.5 * np.exp(-es0[4] * grid.points**2)
    # check norm of density
    assert_almost_equal(
        grid.integrate(dens * 4.0 * np.pi * grid.points**2.0), np.sum(cs0), decimal=6
    )
    # model density is a normalized 1s Gaussian
    model = AtomicGaussianDensity(grid.points, num_s=5, num_p=0, normalize=True)
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens, model, measure=measure, method="slsqp", spherical=True)
    initial_cs = np.array([1., 0., 3., 0., 5.])
    initial_es = np.array([0.1, 1.2, 1., 20., 10.])
    # opt. coeffs
    result = ls.run(initial_cs, es0, True, False, tol=1e-15)
    assert_almost_equal(np.sort(cs0), np.sort(result["coeffs"]), decimal=6)
    assert_almost_equal(np.sort(es0), np.sort(result["exps"]), decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # opt. expons, by re-arranging the exponents and coefficients.
    initial_cs = np.array([0.12, 0.97,  3.67, 1.57, 5.05])
    initial_es = np.array([1.2, 19., 1., 0.1, 10.])
    result = ls.run(initial_cs, initial_es, False, True, tol=1e-20, with_constraint=False)
    assert_almost_equal(np.sort(initial_cs), np.sort(result["coeffs"]), decimal=6)
    assert_almost_equal(np.sort(es0), np.sort(result["exps"]), decimal=1)
    assert_almost_equal(0., result["fun"], decimal=8)


def test_ls_fit_unnormalized_1d_molecular_dens_unnormalized_2s_gaussian():
    r"""Test ScipyFit least-squares against un-normalized Gaussian model."""
    # density is normalized 1s + 1s gaussians
    grid = UniformRadialGrid(200, 0.0, 15.0)
    cs0, es0 = np.array([1.52, 2.67, ]), np.array([0.31, 0.41])
    coords = np.array([[0.], [1.]])
    # compute density on each center
    dens1 = cs0[0] * np.exp(-es0[0] * (grid.points - coords[0])**2.)
    dens2 = cs0[1] * np.exp(-es0[1] * (grid.points - coords[1])**2.)
    # un-normalized 1s + 1s basis functions
    model = MolecularGaussianDensity(grid.points, coords, np.array([[1, 0], [1, 0]]), False)
    # fit total density
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens1 + dens2, model, measure=measure, method="slsqp", spherical=True)
    # initial coeff=1. & expon=1.
    result = ls.run(np.ones(2), np.ones(2), True, True)
    assert_almost_equal(cs0, result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # fit 1s density on center 1
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens1, model, measure=measure, method="slsqp", spherical=True)
    # initial coeff=1. & expon=1.
    result = ls.run(np.ones(2), np.ones(2), True, True)
    assert_almost_equal(np.array([1.52, 0.0]), result["coeffs"], decimal=6)
    assert_almost_equal(es0[0], result["exps"][0], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # fit 1s density on center 2
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens2, model, measure=measure, method="slsqp", spherical=True)
    # initial coeff=1. & expon=1.
    result = ls.run(np.ones(2), np.ones(2), True, True)
    assert_almost_equal(np.array([0.0, 2.67]), result["coeffs"], decimal=6)
    assert_almost_equal(es0[1], result["exps"][1], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)


def test_ls_fit_normalized_1d_molecular_dens_unnormalized_1s_1p_gaussian():
    r"""Test ScipyFit least-squares of 3D 1S, 1P-type Gaussian model."""
    # density is normalized 1s + 1s gaussians
    grid = UniformRadialGrid(200, 0.0, 15.0)
    cs0, es0 = np.array([1.52, 2.67]), np.array([0.31, 0.41])
    coords = np.array([[0.0], [1.0]])
    # compute density of each center
    dens_s = cs0[0] * (es0[0] / np.pi)**1.5 * np.exp(-es0[0] * (grid.points - coords[0])**2.)
    dens_p = cs0[1] * (grid.points - coords[1])**2 * np.exp(-es0[1] * (grid.points - coords[1])**2)
    dens_p *= (2. * es0[1]**2.5 / (3. * np.pi**1.5))
    # normalized 1s + 1p basis functions
    model = MolecularGaussianDensity(grid.points, coords, np.array([[1, 0], [0, 1]]), False)
    expected_cs = cs0 * np.array([(es0[0] / np.pi)**1.5, 2. * es0[1]**2.5 / (3. * np.pi**1.5)])
    # fit total density
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens_s + dens_p, model, measure=measure, method="slsqp", spherical=True)
    # opt. coeffs & expons
    result = ls.run(np.ones(2), np.ones(2), True, True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # opt. coeffs
    result = ls.run(np.array([0.001, 3.671]), es0, True, False)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # opt. expons
    result = ls.run(expected_cs, np.array([2.5, 1.9]), False, True)
    assert_almost_equal(expected_cs, result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # fit 1s density on center 1
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens_s, model, measure=measure, method="slsqp", spherical=True)
    # opt. coeffs & expons
    result = ls.run(np.ones(2), np.ones(2), True, True)
    assert_almost_equal(np.array([expected_cs[0], 0.]), result["coeffs"], decimal=6)
    assert_almost_equal(es0[0], result["exps"][0], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # opt. coeffs
    result = ls.run(np.array([0.05, 5.01]), es0, True, False)
    assert_almost_equal(np.array([expected_cs[0], 0.]), result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
    # fit 1s density on center 1
    measure = SquaredDifference()
    ls = ScipyFit(grid, dens_p, model, measure=measure, method="slsqp", spherical=True)
    result = ls.run(np.array([0.05, 5.01]), es0, True, False, with_constraint=False)
    assert_almost_equal([0., expected_cs[1]], result["coeffs"], decimal=6)
    assert_almost_equal(es0, result["exps"], decimal=6)
    assert_almost_equal(0., result["fun"], decimal=8)
